from musicscore.musicxml.attributes.accidental import Smulf
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.attributes.printstyle import PrintStyle, PrintStyleAlign
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.simple_type import String


class ComplexType(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# TODO: add exception for attributes
class Empty(ComplexType, XMLElement):
    """
    Empty is a type of XMLElement with no children, no text and no attributes.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)

    def add_child(self, child):
        raise Exception('Empty cannot have children.')

    @XMLElement.text.setter
    def text(self, value):
        if value is not None:
            raise Exception('Empty cannot have text.')


class EmptyPlacement(Empty, PrintStyle, Placement):
    """
    The empty-placement type represents an empty element with print-style and placement attributes.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class EmptyPrintStyleAlignId(Empty, PrintStyleAlign, OptionalUniqueId):
    """
    The empty-print-style-align-id type represents an empty element with print-style-align and optional-unique-id 
    attribute groups.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class EmptyPrintObjectStyleAlign(Empty, PrintObject, PrintStyleAlign):
    """
    The empty-print-style-align-object type represents an empty element with print-object and print-style-align
    attribute groups.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class EmptyPlacementSmulf(Empty, Placement, Smulf):
    """
    The empty-placement-smufl type represents an empty element with print-style, placement, and smufl attributes.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class EmptyFont(Empty, Font):
    """
    >The empty-font type represents an empty element with font attributes.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class PlacementText(ComplexType, String, PrintStyle, Placement):
    """
    The placement-text type represents a text element with print-style and placement attribute groups.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class OtherPlacementText(ComplexType, String, PrintStyle, Placement, Smulf):
    """
    The other-placement-text type represents a text element with print-style, placement, and smufl attribute groups.
    This type is used by MusicXML notation extension elements to allow specification of specific SMuFL glyphs without
    needed to add every glyph as a MusicXML element.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, TypeLineType


class LineType(AttributeAbstract):
    """
    The line-type attribute distinguishes between solid, dashed, dotted, and wavy lines.
    """

    def __init__(self, line_type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('line-type', line_type, "TypeLineType")
        TypeLineType


from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeDivisions


class Sound(AttributeAbstract):

    def __init__(self, sound=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('sound', sound, 'TypeYesNo')


class ComplexTypeOffset(ComplexType, TypeDivisions, Sound):
    """
    An offset is represented in terms of divisions, and indicates where the direction will appear relative to the
    current musical location. This affects the visual appearance of the direction. If the sound attribute is "yes", then
    the offset affects playback too. If the sound attribute is "no", then any sound associated with the direction takes
    effect at the current location. The sound attribute is "no" by default for compatibility with earlier versions of
    the MusicXML format. If an element within a direction includes a default-x attribute, the offset value will be
    ignored when determining the appearance of that element
    """

    def __init__(self, tag, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



from musicscore.dtd.dtd import Sequence, GroupReference, Element
from musicscore.musicxml.attributes.document_attributes import DocumentAttributes
from musicscore.musicxml.attributes.measure_attributes import MeasureAttributes
from musicscore.musicxml.attributes.part_attributes import PartAttributes
from musicscore.musicxml.groups.musicdata import MusicData
from musicscore.musicxml.elements.scoreheader import ScoreHeader
from musicscore.musicxml.elements.xml_element import XMLElement


class Part(XMLElement, PartAttributes):
    _DTD = Sequence(
        GroupReference(MusicData)
    )

    def __init__(self, id, *args, **kwargs):
        super().__init__(tag='part', id=id, *args, **kwargs)


class Measure(XMLElement, MeasureAttributes):
    _DTD = Sequence(
        Element(Part, max_occurrence=None)
    )

    def __init__(self, number, *args, **kwargs):
        super().__init__(tag='measure', number=number, *args, **kwargs)


class Score(XMLElement, DocumentAttributes):
    """
    The score-timewise element is the root element for a timewise MusicXML score. It includes a score-header group
    followed by a series of measures with parts inside. The document-attributes attribute group includes the version
    attribute.
    """
    _DTD = Sequence(
        GroupReference(ScoreHeader),
        Element(Measure, max_occurrence=None)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='score-timewise', *args, **kwargs)

    def write(self, path):
        xmlversion = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
        doctype = '<!DOCTYPE score-timewise PUBLIC "-//Recordare//DTD MusicXML {} Timewise//EN" "http://www.musicxml.org/dtds/timewise.dtd">\n'.format(
            self.version)

        if path[-4:] != '.xml':
            path += '.xml'
        output_file = open(path, 'w')
        output_file.write(xmlversion)
        output_file.write(doctype)
        output_file.write(self.to_string())
        output_file.close()



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[0, 4])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_1'
        self.score.write(path=result_path)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[1, 0, 3])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_2'
        self.score.write(path=result_path)

    def test_3(self):
        sf = SimpleFormat(quarter_durations=[4, 0])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_3'
        self.score.write(path=result_path)

    def test_4(self):
        sf = SimpleFormat(quarter_durations=[1.5, 0, 2.5])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_4'
        self.score.write(path=result_path)

    def test_5(self):
        sf = SimpleFormat(quarter_durations=[1.75, 0, 2.25])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_5'
        self.score.write(path=result_path)

    def test_6(self):
        sf = SimpleFormat(quarter_durations=[1.75, 0, 2, 0, 4.25])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_6'
        self.score.write(path=result_path)

    def test_7(self):
        sf = SimpleFormat(quarter_durations=[1.3, 0, 2, 0, 0.7])
        for index, chord in enumerate(sf.chords):
            chord.add_lyric(index + 1)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        # self.score.fill_with_rest()
        # self.score.add_beats()
        # self.score.quantize()
        # self.score.split_not_notatable()
        # self.score.update_tuplets()
        # self.score.substitute_sextoles()
        # self.score.update_types()
        # self.score.update_dots()
        # self.score.chord_to_notes()
        # self.score.update_divisions()
        # self.score.update_accidentals(mode='normal')
        # self.score.update_durations()
        # self.score.finish()
        result_path = path + '_test_7'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from musicscore.dtd.dtd import Sequence, Element, Choice
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.bookmark import ComplexTypeBookmark
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.formattedsymbolid import ComplexTypeFormattedSymbolId
from musicscore.musicxml.types.complextypes.formattedtextid import ComplexTypeFormattedTextId
from musicscore.musicxml.types.complextypes.image import ComplexTypeImage
from musicscore.musicxml.types.complextypes.link import ComplexTypeLink
from musicscore.musicxml.types.simple_type import String


class CreditType(XMLElement, String):
    _TAG = 'credit-type'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Link(ComplexTypeLink):
    _TAG = 'link'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Bookmark(ComplexTypeBookmark):
    _TAG = 'bookmark'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class CreditImage(ComplexTypeImage):
    _TAG = 'credit-image'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class CreditWords(ComplexTypeFormattedTextId):
    _TAG = 'credit-words'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class CreditSymbol(ComplexTypeFormattedSymbolId):
    _TAG = 'credit-symbol'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Page(AttributeAbstract):

    def __init__(self, page=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('page', page, "PositiveInteger")


class ComplexTypeCredit(ComplexType, Page, OptionalUniqueId):
    """The credit type represents the appearance of the title, composer, arranger, lyricist, copyright, dedication, and
    other text, symbols, and graphics that commonly appear on the first page of a score. The credit-words,
    credit-symbol, and credit-image elements are similar to the words, symbol, and image elements for directions.
    However, since the credit is not part of a measure, the default-x and default-y attributes adjust the origin
    relative to the bottom left-hand corner of the page. The enclosure for credit-words and credit-symbol is none
    by default.

    By default, a series of credit-words and credit-symbol elements within a single credit element follow one another
    in sequence visually. Non-positional formatting attributes are carried over from the previous element by default.

    The page attribute for the credit element specifies the page number where the credit should appear. This is an
    integer value that starts with 1 for the first page. Its value is 1 by default. Since credits occur before the
    music, these page numbers do not refer to the page numbering specified by the print element's page-number attribute.

    The credit-type element indicates the purpose behind a credit. Multiple types of data may be combined in a single
    credit, so multiple elements may be used. Standard values include page number, title, subtitle, composer, arranger,
    lyricist, and rights."""
    _DTD = Sequence(
        Element(CreditType, min_occurrence=0, max_occurrence=None),
        Element(Link, min_occurrence=0, max_occurrence=None),
        Element(Bookmark, min_occurrence=0, max_occurrence=None),
        Choice(
            Element(CreditImage),
            Sequence(
                Choice(
                    Element(CreditWords),
                    Element(CreditSymbol)
                ),
                # Sequence(
                #     Element(Link, min_occurrence=0, max_occurrence=None),
                #     Element(Bookmark, min_occurrence=0, max_occurrence=None),
                #     Choice(
                #         Element(CreditWords),
                #         Element(CreditSymbol)
                #     ),
                #     min_occurrence=0,
                #     max_occurrence=None,
                # )
            )

        )
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class TestTreeTimewise(TestCase):
    def setUp(self):
        self.score = TreeScoreTimewise()
        self.score.add_measure()
        self.score.add_part()

    def test_score(self):
        result_path = path + '_test_score'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_add_note(self):
        self.score.add_chord(1, 1, TreeChord(0, 1))
        self.score.add_chord(1, 1, TreeChord(0, quarter_duration=1))
        self.score.add_chord(1, 1, TreeChord(61, quarter_duration=2))
        self.score.finish()
        result_path = path + '_test_add_note'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_add_chord(self):
        self.score.add_chord(1, 1, TreeChord((60, 61), quarter_duration=4))
        self.score.finish()
        # print(self.score.to_string())
        # self.score.write(path=path)

    def test_4(self):
        self.score.tuplet_line_width = 2.4
        sf = SimpleFormat(quarter_durations=[0.2, 0.8])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_4.xml'
        self.score.write(xml_path)
        TestScore().assert_template(result_path=xml_path)



from musicscore.musictree.midi import Midi, C, G, E, B, F, D, A
from musicscore.musicxml.types.complextypes.attributes import Clef
from musicscore.musicxml.types.complextypes.clef import Sign, Line, ClefOctaveChange


class TreeClef(Clef):

    def __init__(self, sign='G', line=2, octave_change=None, optimal_range=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sign = None
        self._line = None
        self._octave_change = None
        self._optimal_range = None

        self.sign = sign
        self.line = line
        self.octave_change = octave_change
        self.optimal_range = optimal_range

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.sign == other.sign and self.line == other.line and self.octave_change == other.octave_change
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def sign(self):
        return self._sign

    @sign.setter
    def sign(self, value):
        if not self._sign:
            self._sign = self.add_child(Sign(value))
        else:
            self._sign.value = value

        self._sign = value

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, value):
        if not self._line:
            self._line = self.add_child(Line(value))
        else:
            self._line.value = value

        self._line = value

    @property
    def octave_change(self):
        return self._octave_change

    @octave_change.setter
    def octave_change(self, value):
        if value:
            if not self._octave_change:
                self._octave_change = self.add_child(ClefOctaveChange(value))
            else:
                self._octave_change.value = value

            self._octave_change = value
        else:
            if self._octave_change:
                self.remove_child(self._octave_change)
                self._octave_change = None

    @property
    def optimal_range(self):
        return self._optimal_range

    @optimal_range.setter
    def optimal_range(self, val):
        if not val:
            val = [None, None]
        if not hasattr(val, '__iter__'):
            raise TypeError('{} must be an iterator.'.format(val))
        if not len(val) == 2:
            raise TypeError('{} must have two elements.'.format(val))
        for index, x in enumerate(val):
            if x and not isinstance(x, Midi):
                val[index] = Midi(x)

        self._optimal_range = val


SUPER_HIGH_TREBLE_CLEF = TreeClef(sign='G', line=2, octave_change=2, optimal_range=[C(7), None])
HIGH_TREBLE_CLEF = TreeClef(sign='G', line=2, octave_change=1, optimal_range=[C(6), None])
TREBLE_CLEF = TreeClef(sign='G', line=2, optimal_range=[G(3), E(6)])
LOW_TREBLE_CLEF = TreeClef(sign='G', line=2, octave_change=-1, optimal_range=[G(2), E(5)])
BASS_CLEF = TreeClef(sign='F', line=4, optimal_range=[B(1), G(4)])
LOW_BASS_CLEF = TreeClef(sign='F', line=4, octave_change=-1, optimal_range=[None, B(1)])
SUPER_LOW_BASS_CLEF = TreeClef(sign='F', line=4, octave_change=-2, optimal_range=[None, G(1)])

ALTO_CLEF = TreeClef(sign='C', line=3, optimal_range=[A(2), F(5)])
TENOR_CLEF = TreeClef(sign='C', line=4, optimal_range=[F(2), D(5)])

PERCUSSION_CLEF = TreeClef('percussion')

ALL_CLEFS = [SUPER_HIGH_TREBLE_CLEF, HIGH_TREBLE_CLEF, TREBLE_CLEF, LOW_TREBLE_CLEF, BASS_CLEF, LOW_BASS_CLEF,
             SUPER_LOW_BASS_CLEF, ALTO_CLEF, TENOR_CLEF]



from musicscore.dtd.dtd import Sequence, Element, GroupReference
from musicscore.musicxml.groups.common import Editorial
from musicscore.musicxml.types.complextypes.complextype import ComplexType, Empty
from musicscore.musicxml.types.complextypes.groupbarline import ComplexTypeGroupBarline
from musicscore.musicxml.types.complextypes.groupname import ComplexTypeGroupName
from musicscore.musicxml.types.complextypes.groupsymbol import ComplexTypeGroupSymbol
from musicscore.musicxml.types.complextypes.namedisplay import ComplexTypeNameDisplay
from musicscore.musicxml.types.simple_type import TypeStartStop, Token


class GroupName(ComplexTypeGroupName):
    _TAG = 'group-name'

    def __init__(self, value='', *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class GroupNameDisplay(ComplexTypeNameDisplay):
    _TAG = 'group-name-display'
    """Formatting specified in the group-name-display element overrides formatting specified in the group-name 
    element."""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class GroupAbbreviation(ComplexTypeGroupName):
    _TAG = 'group-abbreviation'

    def __init__(self, value='', *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class GroupAbbreviationDisplay(ComplexTypeNameDisplay):
    """
    Formatting specified in the group-abbreviation-display element overrides formatting specified in the
    group-abbreviation element.
    """
    _TAG = 'group-abbreviation-display'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class GroupSymbol(ComplexTypeGroupSymbol):
    _TAG = 'group-symbol'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class GroupBarline(ComplexTypeGroupBarline):
    _TAG = 'group-barline'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class GroupTime(Empty):
    """The group-time element indicates that the displayed time signatures should stretch across all parts and
    staves in the group."""

    _TAG = 'group-time'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypePartGroup(ComplexType):
    """
    The part-group element indicates groupings of parts in the score, usually indicated by braces and brackets. Braces
    that are used for multi-staff parts should be defined in the attributes element for that part. The part-group start
    element appears before the first score-part in the group. The part-group stop element appears after the last
    score-part in the group.

    The number attribute is used to distinguish overlapping and nested part-groups, not the sequence of groups. As with
    parts, groups can have a name and abbreviation. Values for the child elements are ignored at the stop of a group.

    A part-group element is not needed for a single multi-staff part. By default, multi-staff parts include a brace
    symbol and (if appropriate given the bar-style) common barlines. The symbol formatting for a multi-staff part can
    be more fully specified using the part-symbol element.
    """
    _DTD = Sequence(
        Element(GroupName, min_occurrence=0),
        Element(GroupNameDisplay, min_occurrence=0),
        Element(GroupAbbreviation, min_occurrence=0),
        Element(GroupAbbreviationDisplay, min_occurrence=0),
        Element(GroupSymbol, min_occurrence=0),
        Element(GroupBarline, min_occurrence=0),
        Element(GroupTime, min_occurrence=0),
        GroupReference(Editorial)

    )
    _ATTRIBUTES = []

    def __init__(self, type, number='1', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = type
        self.number = number

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            Token(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStop(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)



from unittest import TestCase

from musicscore.musicxml.elements.note import Notations
from musicscore.musicxml.types.complextypes.notations import Ornaments
from musicscore.musicxml.types.complextypes.ornaments import Tremolo, Turn
from musicscore.musicxml.types.complextypes.scorepart import PartAbbreviation


class Test(TestCase):
    def test_1(self):
        pa = PartAbbreviation('bla')
        pa.font_size = 10
        result = """<part-abbreviation font-size="10">bla</part-abbreviation>
"""
        self.assertEqual(pa.to_string(), result)



from unittest import TestCase

from musicscore.musicxml.types.complextypes.namedisplay import DisplayText, AccidentalText
from musicscore.musicxml.types.complextypes.partgroup import GroupName, GroupNameDisplay, GroupSymbol, GroupBarline
from musicscore.musicxml.types.complextypes.partlist import PartGroup


class Test(TestCase):
    def test_1(self):
        obj = GroupName()
        expected = """<group-name></group-name>
"""
        result = obj.to_string()
        self.assertEqual(expected, result)

    def test_2(self):
        obj = GroupName('strings')
        expected = """<group-name>strings</group-name>
"""
        result = obj.to_string()
        self.assertEqual(expected, result)

    def test_3(self):
        obj = GroupNameDisplay()
        obj.add_child(DisplayText('strings'))
        obj.add_child(AccidentalText('natural'))
        expected = """<group-name-display>
  <display-text>strings</display-text>
  <accidental-text>natural</accidental-text>
</group-name-display>
"""
        result = obj.to_string()
        self.assertEqual(expected, result)

    def test_4(self):
        obj = GroupNameDisplay()
        obj.add_child(DisplayText('strings'))
        obj.add_child(AccidentalText('natural'))
        expected = """<group-name-display>
  <display-text>strings</display-text>
  <accidental-text>natural</accidental-text>
</group-name-display>
"""
        result = obj.to_string()
        self.assertEqual(expected, result)

    def test_5(self):
        pg = PartGroup(type='start', number='2')
        gnd = pg.add_child(GroupNameDisplay())
        gnd.add_child(DisplayText('strings'))
        pg.add_child(GroupSymbol('bracket'))
        pg.add_child(GroupBarline('yes'))
        expected = """<part-group number="2" type="start">
  <group-name-display>
    <display-text>strings</display-text>
  </group-name-display>
  <group-symbol>bracket</group-symbol>
  <group-barline>yes</group-barline>
</part-group>
"""
        result = pg.to_string()
        self.assertEqual(expected, result)





import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.groups.musicdata import Direction
from musicscore.musicxml.types.complextypes.direction import DirectionType
from musicscore.musicxml.types.complextypes.directiontype import Words
from tests.score_templates.xml_test_score import TestScore
from musicscore.musictree.wordsymbols import SALTANDO

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[1, 1])

        sf.chords[0].add_words('first word')
        sf.chords[1].add_words('second word', font_family='DejaVu Sans', font_size=14, font_weight='bold',
                               font_style='italic')

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[1, 1])

        sf.chords[0].add_words('a')
        sf.chords[1].add_words('b')
        sf.chords[0].add_words('aa', relative_y=-15)

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_2'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        sf = SimpleFormat(quarter_durations=[1.95, 2.05])

        sf.chords[0].add_words('a')
        sf.chords[1].add_words('b')
        sf.chords[0].add_words('aa', relative_y=-15)

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_3'
        self.score.write(path=result_path)
        # TestScore().assert_template(result_path=result_path)

    def test_4(self):
        sf = SimpleFormat(quarter_durations=[2])

        chord = sf.chords[0]
        d = chord.add_child(Direction(placement='above'))
        dt = d.add_child(DirectionType())
        dt.add_child(Words('A', font_size=10))
        dt.add_child(Words('B', font_size=20))

        v = sf.to_stream_voice().add_to_score(self.score)

        result_path = path + '_test_4'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_5(self):
        sf = SimpleFormat(quarter_durations=[2])
        sf.chords[0].add_words(SALTANDO)

        sf.to_stream_voice().add_to_score(self.score)

        result_path = path + '_test_5'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from musicscore.musicxml.types.complextypes.complextype import ComplexType


class ComplexTypeDistance(ComplexType):
    """"""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        raise NotImplementedError()



from musicscore.dtd.dtd import Sequence, Choice, Element
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.types.complextypes.accidentaltext import ComplexTypeAccidentalText
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.formattedtext import ComplexTypeFormattedText


class DisplayText(ComplexTypeFormattedText):
    _TAG = 'display-text'

    def __init__(self, value='', *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class AccidentalText(ComplexTypeAccidentalText):
    _TAG = 'accidental-text'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class ComplexTypeNameDisplay(ComplexType, PrintObject):
    """ The name-display type is used for exact formatting of multi-font text in part and group names to the left of
    the system. The print-object attribute can be used to determine what, if anything, is printed at the start of each
    system. Enclosure for the display-text element is none by default. Language for the display-text element is Italian
    ("it") by default.
    """
    _DTD = Sequence(
        Choice(
            Element(DisplayText),
            Element(AccidentalText),
            min_occurrence=0,
            max_occurrence=None
        )
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class StealTimePrevious(AttributeAbstract):

    def __init__(self, steal_time_previous=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('steal-time-previous', steal_time_previous, 'Percent')


class StealTimeFollowing(AttributeAbstract):

    def __init__(self, steal_time_following=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('steal-time-following', steal_time_following, 'TypePercent')


class MakeTime(AttributeAbstract):

    def __init__(self, make_time=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('make-time', make_time, 'TypeDivisions')


class Slash(AttributeAbstract):

    def __init__(self, slash=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('slash', slash, 'TypeYesNo')



import os

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treeclef import ALTO_CLEF, TREBLE_CLEF
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])
durations = [Fraction(31, 14), Fraction(93, 56), Fraction(31, 28), Fraction(155, 56), Fraction(57, 56)]
clefs = [ALTO_CLEF, None, None, TREBLE_CLEF, ALTO_CLEF]


def get_chords():
    return score.get_measure(1).get_part(1).voices[1].chords


sf = SimpleFormat(quarter_durations=durations)
for chord, clef in zip(sf.chords, clefs):
    if clef:
        chord.add_clef(clef)

score = TreeScoreTimewise()
score.max_division = 7
sf.to_stream_voice().add_to_score(score)
xml_path = path + '_1.xml'
score.write(xml_path)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.id import Id
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import String


class Port(AttributeAbstract):
    def __init__(self, port, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('port', port, "TypeMidi16")


class ComplexTypeMidiDevice(ComplexType, String, Port, Id):
    """The midi-device type corresponds to the DeviceName meta event in Standard MIDI Files. The optional port attribute
    is a number from 1 to 16 that can be used with the unofficial MIDI port (or cable) meta event. Unlike the DeviceName
    meta event, there can be multiple midi-device elements per MusicXML part starting in MusicXML 3.0. The optional id
    attribute refers to the score-instrument assigned to this device. If missing, the device assignment affects all
    score-instrument elements in the score-part."""

    def __init__(self, tag, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



from musicscore.dtd.dtd import Sequence, Element, GroupReference
from musicscore.musicxml.attributes.directive import Directive
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.groups.common import EditorialVoiceDirection, Staff
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.directiontype import ComplexTypeDirectionType
from musicscore.musicxml.types.complextypes.offset import ComplexTypeOffset
from musicscore.musicxml.types.complextypes.sound import ComplexTypeSound


class DirectionType(ComplexTypeDirectionType):
    _TAG = 'direction-type'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Offset(ComplexTypeOffset):
    _TAG = 'offset'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Sound(ComplexTypeSound):
    _TAG = 'sound'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeDirection(ComplexType, Placement, Directive, OptionalUniqueId):
    """A direction is a musical indication that is not necessarily attached to a specific note. Two or more may be
    combined to indicate starts and stops of wedges, dashes, etc. For applications where a specific direction is indeed
    attached to a specific note, the direction element can be associated with the note element that follows it in score
    order that is not in a different voice.

    By default, a series of direction-type elements and a series of child elements of a direction-type within a single
    direction element follow one another in sequence visually. For a series of direction-type children, non-positional
    formatting attributes are carried over from the previous element by default."""

    _DTD = Sequence(
        Element(DirectionType, max_occurrence=None),
        Element(Offset, min_occurrence=0),
        GroupReference(EditorialVoiceDirection),
        GroupReference(Staff, min_occurrence=0),
        Element(Sound, min_occurrence=0)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.groupnametext import GroupNameText
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import String


class ComplexTypeGroupName(ComplexType, String, GroupNameText):
    """
    The group-name type describes the name or abbreviation of a part-group element. Formatting attributes in the
    group-name type are deprecated in Version 2.0 in favor of the new group-name-display and group-abbreviation-display
    elements.
    """

    def __init__(self, tag, value='', *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



from musicscore.musicxml.elements.scoreheader import Defaults
from musicscore.musicxml.groups.layout import PageLayout, StaffLayout, SystemLayout
from musicscore.musicxml.groups.margins import LeftMargin, RightMargin, TopMargin, BottomMargin
from musicscore.musicxml.types.complextypes.defaults import Scaling
from musicscore.musicxml.types.complextypes.pagelayout import PageMargins, PageHeight, PageWidth
from musicscore.musicxml.types.complextypes.scaling import Millimeters, Tenths
from musicscore.musicxml.types.complextypes.stafflayout import StaffDistance
from musicscore.musicxml.types.complextypes.systemlayout import SystemDistance, SystemMargins, TopSystemDistance


class TreePageStyle(object):
    sizes = {'A4': (210, 297), 'A3': (297, 420)}

    def __init__(self, score, scale=1, size='A4', orientation='portrait', left_margin=111, right_margin=55,
                 top_margin=83,
                 bottom_margin=83, system_distance=None, system_left_margin=None, system_right_margin=None,
                 top_system_distance=None, staff_distance=None):

        self._score = None
        self.score = score
        self.tenth = 40

        self._size = None
        self._orientation = 'portrait'

        self._defaults = None
        self._scaling = None

        self._page_layout = None
        self._page_height = None
        self._page_width = None
        self._page_margins = None
        self.previous_page_width_value = None
        self.previous_page_height_value = None

        self._left_margin = None
        self._right_margin = None
        self._top_margin = None
        self._bottom_margin = None

        self._system_layout = None
        self._system_margins = None
        self._system_distance = None
        self._staff_layout = None
        self._staff_distance = None
        self._system_left_margin = None
        self._system_right_margin = None
        self._top_system_distance = None

        self.scale = scale
        self.size = size
        self.orientation = orientation

        self.left_margin = left_margin
        self.right_margin = right_margin
        self.top_margin = top_margin
        self.bottom_margin = bottom_margin
        self.staff_distance = staff_distance

        self.system_distance = system_distance
        self.system_left_margin = system_left_margin
        self.system_right_margin = system_right_margin
        self.top_system_distance = top_system_distance

    def millimeters_to_tenth(self, m):
        return round(m / self.millimeters * self.tenth)

    @property
    def millimeters(self):
        return self.scale * 7.2319

    def _add_defaults(self):
        self._defaults = self.score.add_child(Defaults())

    def _add_scaling(self):
        if not self._defaults:
            self._add_defaults()
        self._scaling = self._defaults.add_child(Scaling())
        self._scaling.add_child(Millimeters(self.millimeters))
        self._scaling.add_child(Tenths(self.tenth))

    def _add_page_layout(self):
        if not self._scaling:
            self._add_scaling()

        self._page_layout = self._defaults.add_child(PageLayout())

    def _add_staff_layout(self):
        if not self._scaling:
            self._add_scaling()

        self._staff_layout = self._defaults.add_child(StaffLayout())

    def _add_system_layout(self):
        if not self._scaling:
            self._add_scaling()

        self._system_layout = self._defaults.add_child(SystemLayout())
        self._system_distance = self._system_layout.add_child(SystemDistance(100))
        self._system_margins = self._system_layout.add_child(SystemMargins())
        self._system_left_margin = self._system_margins.add_child(LeftMargin(0))
        self._system_right_margin = self._system_margins.add_child(RightMargin(0))
        self._top_system_distance = self._system_layout.add_child(TopSystemDistance(150))

    def _add_page_margins(self):
        if not self._page_layout:
            self._add_page_layout()

        self._page_margins = self._page_layout.add_child(PageMargins(type_='both'))

    @property
    def score(self):
        '''
        system_layout = defaults.add_child(SystemLayout())
        system_margins = system_layout.add_child(SystemMargins())
        system_margins.add_child(LeftMargin(0))
        system_margins.add_child(RightMargin(0))
        system_layout.add_child(SystemDistance(121))
        system_layout.add_child(TopSystemDistance(300))
        '''
        return self._score

    @score.setter
    def score(self, value):
        self._score = value

    @property
    def page_height(self):
        return self._page_height

    @page_height.setter
    def page_height(self, value):
        value = self.millimeters_to_tenth(value)
        if self.page_height:
            if self.page_height.value != value:
                self.previous_page_height_value = self.page_height.value
                self.page_height.value = value
                self.score.recalculate_y()
        else:
            if not self._page_layout:
                self._add_page_layout()
            self._page_height = self._page_layout.add_child(PageHeight(value))

    @property
    def page_width(self):
        return self._page_width

    @page_width.setter
    def page_width(self, value):
        value = self.millimeters_to_tenth(value)
        if self.page_width:
            if self.page_width.value != value:
                self.previous_page_width_value = self.page_width.value
                self.page_width.value = value
                self.score.recalculate_x()
        else:
            if not self._page_layout:
                self._add_page_layout()
            self._page_width = self._page_layout.add_child(PageWidth(value))

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, value):
        if value == 'landscape':
            self.page_height, self.page_width = self.size[1], self.size[0]
        elif value == 'portrait':
            self.page_height, self.page_width = self.size[0], self.size[1]
        else:
            raise ValueError()

        self._orientation = value

    @property
    def left_margin(self):
        return self._left_margin

    @left_margin.setter
    def left_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.left_margin:
            self.left_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._left_margin = self._page_margins.add_child(LeftMargin(value))

    @property
    def right_margin(self):
        return self._right_margin

    @right_margin.setter
    def right_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.right_margin:
            self.right_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._right_margin = self._page_margins.add_child(RightMargin(value))

    @property
    def top_margin(self):
        return self._top_margin

    @top_margin.setter
    def top_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.top_margin:
            self.top_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._top_margin = self._page_margins.add_child(TopMargin(value))

    @property
    def bottom_margin(self):
        return self._bottom_margin

    @bottom_margin.setter
    def bottom_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.bottom_margin:
            self.bottom_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._bottom_margin = self._page_margins.add_child(BottomMargin(value))

    @property
    def staff_distance(self):
        return self._staff_distance

    @staff_distance.setter
    def staff_distance(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if self.staff_distance:
                self.staff_distance.value = value
            else:
                if not self._staff_layout:
                    self._add_staff_layout()
                self._staff_distance = self._staff_layout.add_child(StaffDistance(value))

    @property
    def system_distance(self):
        return self._system_distance

    @system_distance.setter
    def system_distance(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.system_distance.value = value

    @property
    def system_left_margin(self):
        return self._system_left_margin

    @system_left_margin.setter
    def system_left_margin(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.system_left_margin.value = value

    @property
    def system_right_margin(self):
        return self._system_right_margin

    @system_right_margin.setter
    def system_right_margin(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.system_right_margin.value = value

    @property
    def top_system_distance(self):
        return self._top_system_distance

    @top_system_distance.setter
    def top_system_distance(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.top_system_distance.value = value

    @property
    def size(self):
        (w, h) = self.sizes[self._size]
        # h = self.millimeters_to_tenth(h)
        # w = self.millimeters_to_tenth(w)
        return h, w

    @size.setter
    def size(self, value):
        self._size = value
        self.orientation = self._orientation
        # self.page_height = self.size[0]
        # self.page_width = self.size[1]

    @property
    def page_size(self):
        return self.size

    @page_size.setter
    def page_size(self, value):
        self.size = value



from musicscore.musicxml.attributes.printstyle import Color
from musicscore.musicxml.elements.xml_element import XMLElement
from unittest import TestCase


class Colored(XMLElement, Color):
    """"""
    # _ATTRIBUTES = ['color']

    def __init__(self, *args, **kwargs):
        super().__init__(tag='color_test', *args, **kwargs)


class TestColor(TestCase):
    def setUp(self):
        self.colored = Colored()

    def test_color(self):
        self.colored.color = '#800080'
        result = '''<color_test color="#800080"/>
'''
        self.assertEqual(self.colored.to_string(), result)
        self.colored.color = '#40800080'
        self.colored.text = 'bla'
        result = '''<color_test color="#40800080">bla</color_test>
'''
        self.assertEqual(self.colored.to_string(), result)
        with self.assertRaises(ValueError):
            self.colored.color = 'b0l0a080'



from musicscore.dtd.dtd import GroupReference, Sequence
from musicscore.musicxml.groups.margins import LeftRightMargins
from musicscore.musicxml.types.complextypes.complextype import ComplexType


class ComplexTypeSystemMargins(ComplexType):
    """
    System margins are relative to the page margins. Positive values indent and negative values reduce the margin size.
    """

    _DTD = Sequence(
        GroupReference(LeftRightMargins)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeFermataShape, TypeUprightInverted


class ComplexTypeFermata(ComplexType, TypeFermataShape, PrintStyle, OptionalUniqueId):
    """The fermata text content represents the shape of the fermata sign. An empty fermata element represents a normal
    fermata. The fermata type is upright if not specified."""

    def __init__(self, tag, value, type=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)
        self.type = type

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeUprightInverted(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)



from unittest import TestCase

from musicscore.musicxml.elements.fullnote import Pitch
from musicscore.musicxml.elements.note import Note, Duration, Notations
from musicscore.musicxml.types.complextypes.hole import HoleClosed

from musicscore.musicxml.types.complextypes.notations import Technical
from musicscore.musicxml.types.complextypes.technical import Hole, SnapPizzicato


class Test(TestCase):
    def setUp(self):
        self.note = Note()
        self.note.add_child(Duration())
        self.note.add_child(Pitch())

    def test_1(self):
        notations = self.note.add_child(Notations())
        technical = notations.add_child(Technical())
        hole = technical.add_child(Hole())
        hole.add_child(HoleClosed(value='no'))
        technical.add_child(SnapPizzicato(relative_y=20))

        result = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <notations>
    <technical>
      <hole>
        <hole-closed>no</hole-closed>
      </hole>
      <snap-pizzicato relative-y="20"/>
    </technical>
  </notations>
</note>
"""
        self.assertEqual(self.note.to_string(), result)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.attributes.attribute_example import AttributeExample


class XMLExampleChild1(XMLElement):
    """
    some documentation
    """

    def __init__(self, *args, **kwargs):
        super().__init__(tag='example-child-1', *args, **kwargs)


class XMLExampleChild(XMLElement):
    """
    some documentation
    """

    def __init__(self, *args, **kwargs):
        super().__init__(tag='example-child-2', *args, **kwargs)


class XMLExample(XMLElement, AttributeExample):
    """
    some documentation
    """
    _ATTRIBUTES = ['attribute-example']
    _DTD = Sequence(
        Element(XMLExampleChild1, min_occurrence=0),
        Element(XMLExampleChild, min_occurrence=0, max_occurrence=None)

    )

    def __init__(self, attribute_example=None, *args, **kwargs):
        super().__init__(tag='example', attribute_example=attribute_example, *args, **kwargs)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat, TreeChord, Midi
from musicscore.musictree.treeinstruments import NaturalHorn
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.horn = NaturalHorn()
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        harmonics = SimpleFormat()
        for i in range(1, 17):
            partial_midi = Midi(round(self.horn.get_partial_midi_value(i) * 2) / 2)
            if i % 7 == 0:
                partial_midi.accidental_mode = 'flat'
            if i % 11 == 0:
                partial_midi.accidental_mode = 'sharp'
            if i % 13 == 0:
                partial_midi.accidental_mode = 'flat'
            chord = TreeChord(midis=partial_midi)

            chord.add_words(i)
            harmonics.add_chord(chord)

        harmonics.to_stream_voice().add_to_score(self.score, part_number=2)

        harmonics.transpose(self.horn.transposition)
        harmonics.to_stream_voice().add_to_score(self.score, part_number=1)

        self.score.accidental_mode = 'modern'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)



from unittest import TestCase

from musicscore.dtd.dtd import Sequence, GroupReference, Choice, Element
from musicscore.musicxml.elements.fullnote import Chord
from musicscore.musicxml.elements.note import Duration, Grace
from musicscore.musicxml.elements.xml_element import XMLElement

TestGroup = Sequence(
    Element(Grace),
    Element(Chord)
)


class Foo(XMLElement):
    _DTD = Sequence(
        Choice(
            Sequence(
                GroupReference(TestGroup),
                Element(Duration)),
            Sequence(
                GroupReference(TestGroup),
            )
        )

    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='foo', *args, **kwargs)


class TestGroupReference(TestCase):
    def setUp(self):
        self.foo = Foo()

    def test_expand(self):
        self.foo.add_child(Chord())
        self.foo.add_child(Grace())
        self.assertEqual([type(child).__name__ for child in self.foo.get_children()], ['Grace', 'Chord'])



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat, TreeChord
from musicscore.musictree.treechordflags2 import FingerTremoloFlag2
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        durations = [2, 2, 2]
        midis = [60, 62, 63]
        sf = SimpleFormat(quarter_durations=durations, midis=midis)
        for chord in sf.chords:
            chord.add_flag(FingerTremoloFlag2(tremolo_chord=TreeChord(midis=[67])))

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        xml_path = path + '_test_2.xml'
        durations = [1.5, 1.5, 1.5]
        midis = [60, 62, 63]
        sf = SimpleFormat(quarter_durations=durations, midis=midis)
        for chord in sf.chords:
            chord.add_flag(FingerTremoloFlag2(tremolo_chord=TreeChord(midis=[67])))

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_3(self):
        xml_path = path + '_test_3.xml'
        durations = [0.25]
        midis = [60]
        sf = SimpleFormat(quarter_durations=durations, midis=midis)
        for chord in sf.chords:
            chord.add_flag(FingerTremoloFlag2(tremolo_chord=TreeChord(midis=[67])))

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)

    def test_4(self):
        xml_path = path + '_test_4.xml'
        durations = [0.2, 0.8, 1]
        midis = [60, 62, 63]
        sf = SimpleFormat(quarter_durations=durations, midis=midis)
        for chord in sf.chords:
            chord.add_flag(FingerTremoloFlag2(tremolo_chord=TreeChord(midis=[67])))

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)




from musicscore.dtd.dtd import Element, Sequence, Choice, GroupReference, ChildTypeDTDConflict, \
    ChildOccurrenceDTDConflict, ChildIsNotOptional
from musicscore.musicxml.groups.common import Editorial
from musicscore.musicxml.elements.fullnote import FullNote, Pitch, Unpitched, Rest, Alter
from musicscore.musicxml.groups.musicdata import Backup, Forward, Direction, Attributes, Sound, Barline, Link, \
    Bookmark
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.attributes import TimeSignature, Interchangeable, SenzaMisura, Beats, BeatType, Time
from musicscore.musicxml.types.complextypes.lyric import Elision, EndLine, EndParagraph, Syllabic, Humming, Laughing, \
    Extend, Text
from unittest import TestCase

from musicscore.musicxml.elements.note import Grace, Play, Lyric, Notations, Beam, Staff, NoteheadText, Notehead, Stem, \
    TimeModification, Accidental, Dot, Type, EditorialVoice, Instrument, Tie, DurationGroup, Cue, Duration, Note

el = Element(Elision)

seq1 = Sequence(
    Element(Syllabic),
    Element(Elision)
)

seq2 = Sequence(
    Element(Syllabic),
    Element(Elision),
    Sequence(
        Element(EndLine),
        Element(EndParagraph)
    )
)

ch1 = Choice(
    Element(Syllabic),
    Element(Elision),
    Element(EndLine)
)

ch2 = Sequence(
    Choice(
        Element(Syllabic),
        Element(Elision)
    ),
    Element(EndLine)
)

ch3 = Choice(
    Sequence(
        Choice(
            Sequence(
                Element(Syllabic),
                Element(Elision)
            ),
            Element(EndParagraph)
        ),
        Element(EndLine)
    ),
    Element(EndLine)
)

lyric_dtd = Sequence(
    Choice(
        Sequence(
            Element(Syllabic, min_occurrence=0),
            Element(Text),
            Sequence(
                Sequence(
                    Element(Elision),
                    Element(Syllabic, min_occurrence=0),
                    min_occurrence=0
                ),
                Element(Text)
                ,
                min_occurrence=0,
                max_occurrence=None
            ),
            Element(Extend, min_occurrence=0)
        ),
        Element(Extend),
        Element(Laughing),
        Element(Humming)
    ),
    Element(EndLine, min_occurrence=0),
    Element(EndParagraph, min_occurrence=0),
    GroupReference(Editorial)
)

note_dtd = Sequence(
    Choice(
        Sequence(
            Element(Grace),
            Choice(
                Sequence(
                    GroupReference(FullNote)
                ),
                Sequence(
                    GroupReference(FullNote),
                    Element(Tie, 0, 2)
                ),
                Sequence(
                    Element(Cue),
                    GroupReference(FullNote)
                )
            )
        ),
        Sequence(
            Element(Cue),
            GroupReference(FullNote),
            GroupReference(DurationGroup)
        ),
        Sequence(
            GroupReference(FullNote),
            GroupReference(DurationGroup),
            Element(Tie, 0, 2)
        )
    ),
    Element(Instrument, 0),
    GroupReference(EditorialVoice, 0),
    Element(Type, 0),
    Element(Dot, 0, None),
    Element(Accidental, 0),
    Element(TimeModification, 0, None),
    Element(Stem, 0),
    Element(Notehead, 0),
    Element(NoteheadText, 0),
    GroupReference(Staff, 0),
    Element(Beam, 0, 8),
    Element(Notations, 0, None),
    Element(Lyric, 0, None),
    Element(Play, 0)
)

MusicData = Sequence(
    Choice(
        Element(Note),
        Element(Backup),
        Element(Forward),
        Element(Direction),
        Element(Attributes),
        Element(Sound),
        Element(Barline),
        Element(Link),
        Element(Bookmark),
        min_occurrence=0,
        max_occurrence=None
    )
)


class TestXML(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='test-xml', *args, **kwargs)


class Test(TestCase):

    def test_dtd_check_type(self):
        # dtd = seq2.__deepcopy__()
        TestXML._DTD = seq2
        test = TestXML()

        test.add_xml_child(Elision('bla'))
        test.add_xml_child(Syllabic('single'))
        with self.assertRaises(ChildTypeDTDConflict):
            test.add_xml_child(Lyric())

        TestXML._DTD = ch2
        test = TestXML()

        # dtd = ch2.__deepcopy__()
        test.add_xml_child(EndLine())
        test.add_xml_child(Syllabic('single'))
        with self.assertRaises(ChildTypeDTDConflict):
            test.add_xml_child(Elision('bla'))

    # def test_dtd_check_max_occurrence(self):
    #     dtd = seq2.__deepcopy__()
    #     dtd.add_xml_child(Elision('bla'))
    #     # print(dtd.get_current_xml_children())
    #     with self.assertRaises(ChildOccurrenceDTDConflict):
    #         dtd.add_xml_child(Elision('bla'))
    #
    #     dtd = ch2.__deepcopy__()
    #     dtd.add_xml_child(EndLine())
    #     dtd.add_xml_child(Syllabic('single'))
    #     with self.assertRaises(ChildOccurrenceDTDConflict):
    #         dtd.add_xml_child(Syllabic('single'))


class TestNoteDtd(TestCase):
    def setUp(self):
        self.note = Note()
        # self.dtd = note_dtd.__deepcopy__()
        self.note.add_xml_child(Instrument())
        self.note.add_xml_child(Rest())
        self.note.add_xml_child(Duration())
        self.note.add_xml_child(Type('quarter'))

    def test_dtd_note_1(self):
        with self.assertRaises(ChildTypeDTDConflict):
            self.note.add_xml_child(Unpitched())

    def test_dtd_note_2(self):
        with self.assertRaises(ChildOccurrenceDTDConflict):
            self.note.add_xml_child(Instrument())

    def test_dtd_note_3(self):
        self.note.add_xml_child(Tie())
        self.note.add_xml_child(Tie())

        result = ['Rest', 'Duration', 'Tie', 'Tie', 'Instrument', 'Type']
        self.assertEqual([type(xml_child).__name__ for xml_child in self.note.current_children], result)

        with self.assertRaises(ChildOccurrenceDTDConflict):
            self.note.add_xml_child(Tie())

    def test_dtd_pitch(self):
        p = Pitch()
        result = ['Step', 'Octave']
        self.assertEqual([type(child).__name__ for child in p.get_children()], result)
        p.add_child(Alter(2))
        result = ['Step', 'Alter', 'Octave']
        self.assertEqual([type(child).__name__ for child in p.get_children()], result)
        # p.close()
        # print(p.to_string())


class TestMusicData(TestCase):
    def setUp(self):
        TestXML._DTD = MusicData
        self.md = TestXML()

    def test_music_data_dtd(self):
        self.md.add_xml_child(Forward())
        self.md.add_xml_child(Bookmark())
        self.md.add_xml_child(Forward())
        self.md.add_xml_child(Direction())
        result = ['Forward', 'Bookmark', 'Forward', 'Direction']
        self.assertEqual([type(xml_child).__name__ for xml_child in self.md.current_children], result)


time_dtd = Choice(
    Sequence(
        GroupReference(TimeSignature, max_occurrence=None),
        Element(Interchangeable, min_occurrence=0)
    ),
    Element(SenzaMisura)
)


class TestTimeDtd(TestCase):
    def setUp(self):
        self.time = Time()

    def test_time(self):
        self.time.add_child(BeatType(4))
        self.time.add_child(Beats(4))
        self.time.add_child(BeatType(2))
        self.time.add_child(BeatType(16))
        self.time.add_child(Beats(5))
        result = [('Beats', 4), ('BeatType', 4), ('Beats', 5), ('BeatType', 2), ('BeatType', 16)]
        self.assertEqual([(type(child).__name__, child.value) for child in self.time.current_children], result)
        with self.assertRaises(ChildIsNotOptional):
            self.time.close_dtd()



from musicscore.musicxml.attributes.bendsound import BendSound
from musicscore.musicxml.attributes.dahsedformatting import DashedFormatting
from musicscore.musicxml.attributes.linetype import LineType
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import Empty
from musicscore.musicxml.types.simple_type import TypeStartStop, TypeNumberLevel


class ComplexTypeSlide(Empty, LineType, DashedFormatting, PrintStyle, BendSound, OptionalUniqueId):
    """Glissando and slide types both indicate rapidly moving from one pitch to the other so that individual notes are
    not discerned. The distinction is similar to that between NIFF's glissando and portamento elements. A slide is
    continuous between two notes and defaults to a solid line. The optional text for a is printed alongside the line."""

    def __init__(self, tag, type, number=1, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.type = type
        self.number = number

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStop(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeNumberLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)



from unittest import TestCase
from musicscore.musicxml.elements.fullnote import Step, Alter, Octave, Pitch


class TestPitch(TestCase):
    def setUp(self):
        self.pitch = Pitch()

    def test_to_string(self):
        result = '<pitch>\n  <step>C</step>\n  <octave>4</octave>\n</pitch>\n'
        self.assertEqual(self.pitch.to_string(), result)

    def test_arguments(self):
        pitch = Pitch(Step('D'), octave=Octave(7), alter=Alter(1.5))
        result = '<pitch>\n  <step>D</step>\n  <alter>1.5</alter>\n  <octave>7</octave>\n</pitch>\n'
        self.assertEqual(pitch.to_string(), result)

    def test_change_alter_1(self):
        self.pitch.alter = Alter(-1)
        result = '<pitch>\n  <step>C</step>\n  <alter>-1</alter>\n  <octave>4</octave>\n</pitch>\n'
        self.assertEqual(self.pitch.to_string(), result)

    def test_change_alter_2(self):
        self.pitch.alter = Alter(-1)
        self.pitch.alter = Alter(1)
        result = '<pitch>\n  <step>C</step>\n  <alter>1</alter>\n  <octave>4</octave>\n</pitch>\n'
        self.assertEqual(self.pitch.to_string(), result)

    def test_set_children(self):
        pitch = Pitch(step='E')
        result = '<pitch>\n  <step>E</step>\n  <octave>4</octave>\n</pitch>\n'
        self.assertEqual(pitch.to_string(), result)



import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
path = str(os.path.abspath(__file__).split('.')[0])
xml_path = path + '.xml'
score = TreeScoreTimewise()
quarter_durations = [1, 1-0.01, 0.01]
sf = SimpleFormat(quarter_durations=quarter_durations)
score.set_time_signatures(round(sum(quarter_durations)))
sf.to_stream_voice().add_to_score(score)
#
# score.update_measures()
# score.fill_with_rest()
# # score.preliminary_adjoin_rests()
# # score.add_beats()
# # score.quantize()
# print([float(chord.quarter_duration) for chord in score.get_measure(1).get_part(1).chords])
score.write(xml_path)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, TypeStaffLine
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeClefSign, Integer


class Sign(XMLElement, TypeClefSign):
    """
    The sign element represents the clef symbol.
    """
    _TAG = 'sign'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Line(XMLElement, TypeStaffLine):
    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='line', value=value, *args, **kwargs)


class ClefOctaveChange(XMLElement, Integer):
    """
    The clef-octave-change element is used for transposing clefs. A treble clef for tenors would have a value of -1.
    """
    _TAG = 'clef-octave-change'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


'''
        <xs:attribute name="number" type="staff-number"/>
        <xs:attribute name="additional" type="yes-no"/>
        <xs:attribute name="size" type="symbol-size"/>
        <xs:attribute name="after-barline" type="yes-no"/>
        <xs:attributeGroup ref="print-style"/>
        <xs:attributeGroup ref="print-object"/>
        <xs:attributeGroup ref="optional-unique-id"/>
'''


class AfterBarline(AttributeAbstract):
    """"""

    def __init__(self, after_barline=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('after-barline', after_barline, "TypeYesNo")


class ComplexTypeClef(ComplexType, PrintStyle, PrintObject, OptionalUniqueId, AfterBarline):
    """
    Clefs are represented by a combination of sign, line, and clef-octave-change elements. The optional number attribute
    refers to staff numbers within the part. A value of 1 is assumed if not present. Sometimes clefs are added to the
    staff in non-standard line positions, either to indicate cue passages, or when there are multiple clefs present
    simultaneously on one staff. In this situation, the additional attribute is set to "yes" and the line value is
    ignored. The size attribute is used for clefs where the additional attribute is "yes". It is typically used to
    indicate cue clefs.

    Sometimes clefs at the start of a measure need to appear after the barline rather than before, as for cues or for
    use after a repeated section. The after-barline attribute is set to "yes" in this situation. The attribute is
    ignored for mid-measure clefs.

    Clefs appear at the start of each system unless the print-object attribute has been set to "no" or the additional
    attribute has been set to "yes".
    """

    _DTD = Sequence(
        Element(Sign),
        Element(Line, min_occurrence=0),
        Element(ClefOctaveChange, min_occurrence=0)
    )



from unittest import TestCase

from quicktions import Fraction

from musicscore.musictree.midi import Midi
from musicscore.musictree.treebeat import TreeBeat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart


class TestTreeChord(TestCase):
    def setUp(self):
        self.chord = TreeChord((60, 61), quarter_duration=2)

    def test_chord(self):
        for tree_note in self.chord._notes:
            tree_note.update_duration(divisions=1)
            # print(tree_note.to_string())

    def test_rest(self):
        with self.assertRaises(ValueError):
            TreeChord((0, 61), quarter_duration=2)

        chord = TreeChord(0, quarter_duration=2)
        tree_note = chord._notes[0]
        tree_note.update_duration(divisions=1)

        result = '''<note>
  <rest/>
  <duration>2</duration>
</note>
'''
        self.assertEqual(tree_note.to_string(), result)

    def test_chord_midi(self):
        chord = TreeChord(Midi(63, accidental_mode='sharp'), quarter_duration=2)
        tree_note = chord._notes[0]
        tree_note.update_duration(divisions=1)

        result = '''<note>
  <pitch>
    <step>D</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
</note>
'''
        self.assertEqual(tree_note.to_string(), result)

    def test_split_chord(self):
        b = TreeBeat()
        p = TreePart('one')
        chord = TreeChord((60, 62), quarter_duration=1)
        m = TreeMeasure()
        m.add_child(p)
        p.add_chord(chord)
        b.add_chord(chord)
        split = chord.split([1, 0.5, 3])

        result = [Fraction(2, 9), Fraction(1, 9), Fraction(2, 3)]
        self.assertEqual([chord.quarter_duration for chord in split], result)



from unittest import TestCase

from musicscore.musicxml.elements import timewise
from musicscore.musicxml.elements.scoreheader import PartList, Defaults
from musicscore.musicxml.types.complextypes.appearance import LineWidth
from musicscore.musicxml.types.complextypes.defaults import Appearance
from musicscore.musicxml.types.complextypes.partlist import ScorePart
from musicscore.musicxml.types.complextypes.scorepart import PartName


class Test(TestCase):
    def setUp(self) -> None:
        self.score = timewise.Score()
        pl = self.score.add_child(PartList())
        sp = pl.add_child(ScorePart(id='p1'))
        sp.add_child(PartName(name='none'))
        m = self.score.add_child(timewise.Measure(number=1))
        m.add_child(timewise.Part(id='p1'))

    def test_1(self):
        d = self.score.add_child(Defaults())
        a = d.add_child(Appearance())
        a.add_child(LineWidth(type='tuplet bracket', value=2.4))
        result = """<score-timewise version="1.0">
  <defaults>
    <appearance>
      <line-width type="tuplet bracket">2.4</line-width>
    </appearance>
  </defaults>
  <part-list>
    <score-part id="p1">
      <part-name>none</part-name>
    </score-part>
  </part-list>
  <measure number="1">
    <part id="p1"/>
  </measure>
</score-timewise>
"""
        self.assertEqual(self.score.to_string(), result)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class LineLength(AttributeAbstract):
    """
    The line-length attribute distinguishes between different line lengths for doit, falloff, plop, and scoop
    articulations.
    """

    def __init__(self, line_length=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('line-length', line_length, "TypeLineLength")


from musicscore.dtd.dtd import Element, GroupReference, Sequence, Choice
from musicscore.musicxml.attributes.accidental import Cautionary, Editorial, Smulf
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.grace_attributes import StealTimePrevious, StealTimeFollowing, MakeTime, Slash
from musicscore.musicxml.attributes.leveldisplay import LevelDisplay
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.position import XPosition
from musicscore.musicxml.attributes.printout import Printout
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.elements.fullnote import FullNote
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.groups.common import EditorialVoice, Staff
from musicscore.musicxml.types.complextypes.beam import ComplexTypeBeam
from musicscore.musicxml.types.complextypes.complextype import EmptyPlacement, ComplexType
from musicscore.musicxml.types.complextypes.lyric import ComplexTypeLyric
from musicscore.musicxml.types.complextypes.notations import ComplexTypeNotations
from musicscore.musicxml.types.complextypes.notehead import ComplexTypeNotehead
from musicscore.musicxml.types.complextypes.notetype import ComplexTypeNoteType
from musicscore.musicxml.types.complextypes.stem import ComplexTypeStem
from musicscore.musicxml.types.complextypes.tie import ComplexTypeTie
from musicscore.musicxml.types.complextypes.timemodification import ComplexTypeTimeModification
from musicscore.musicxml.types.simple_type import TypePositiveDivisions, TypeAccidentalValue


class Grace(XMLElement, StealTimePrevious, StealTimeFollowing, MakeTime, Slash):
    """
    documentation>The grace type indicates the presence of a grace note. The slash attribute for a grace note is yes
    for slashed eighth notes. The other grace note attributes come from MuseData sound suggestions. The
    steal-time-previous attribute indicates the percentage of time to steal from the previous note for the grace note.
    The steal-time-following attribute indicates the percentage of time to steal from the following note for the grace
    note, as for appoggiaturas. The make-time attribute indicates to make time, not steal time; the units are in
    real-time divisions for the grace note.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(tag='grace', *args, **kwargs)


# type="semitones"


class Tie(ComplexTypeTie):
    """
    The tie element indicates that a tie begins or ends with this note. If the tie element applies only particular times
    through a repeat, the time-only attribute indicates which times to apply it. The tie element indicates sound; the
    tied element indicates notation.</xs:documentation>
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Cue(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='cue', *args, **kwargs)


class Duration(XMLElement, TypePositiveDivisions):
    """
    Duration is a positive number specified in division units. This is the intended duration vs. notated duration
    (for instance, swing eighths vs. even eighths, or differences in dotted notes in Baroque-era music). Differences
    in duration specific to an interpretation or performance should use the note element's attack and release
    attributes
    """

    def __init__(self, value=1, *args, **kwargs):
        super().__init__(tag='duration', value=value, *args, **kwargs)


"""The duration element is defined within a group due to its uses within the note, figure-bass, backup, and
forward elements.
"""
DurationGroup = Sequence(

    Element(Duration)
)


class Instrument(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='instrument', *args, **kwargs)


class Type(ComplexTypeNoteType):
    """"""
    _TAG = 'type'

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Dot(EmptyPlacement):
    """
    One dot element is used for each dot of prolongation. The placement element is used to specify whether the dot
    should appear above or below the staff line. It is ignored for notes that appear on a staff space.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(tag='dot', *args, **kwargs)


class Accidental(ComplexType, TypeAccidentalValue, Cautionary, Editorial, LevelDisplay, PrintStyle, Smulf):
    """
    The accidental type represents actual notated accidentals. Editorial and cautionary indications are indicated by 
    attributes. Values for these attributes are "no" if not present. Specific graphic display such as parentheses, 
    brackets, and size are controlled by the level-display attribute group
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='accidental', value=value, *args, **kwargs)


class TimeModification(ComplexTypeTimeModification):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Stem(ComplexTypeStem):
    """"""
    _TAG = 'stem'

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Notehead(ComplexTypeNotehead):
    _TAG = 'notehead'
    """"""

    def __init__(self, value='normal', *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class NoteheadText(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='notehead-text', *args, **kwargs)
        raise NotImplementedError()


class Beam(ComplexTypeBeam):
    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class Notations(ComplexTypeNotations):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Lyric(ComplexTypeLyric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Play(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='play', *args, **kwargs)
        raise NotImplementedError()


"""
        <xs:attributeGroup ref="x-position"/>
        <xs:attributeGroup ref="font"/>
        <xs:attributeGroup ref="color"/>
        <xs:attributeGroup ref="printout"/>
        <xs:attribute name="print-leger" type="yes-no"/>
        <xs:attribute name="dynamics" type="non-negative-decimal"/>
        <xs:attribute name="end-dynamics" type="non-negative-decimal"/>
        <xs:attribute name="attack" type="divisions"/>
        <xs:attribute name="release" type="divisions"/>
        <xs:attribute name="time-only" type="time-only"/>
        <xs:attribute name="pizzicato" type="yes-no"/>
        <xs:attributeGroup ref="optional-unique-id"/>"""


class Note(XMLElement, XPosition, Font, Color, Printout, OptionalUniqueId):
    """
    Notes are the most common type of MusicXML data. The MusicXML format keeps the MuseData distinction between
    elements used for sound information and elements used for notation information (e.g., tie is used for sound,
    tied for notation). Thus grace notes do not have a duration element. Cue notes have a duration element, as do
    forward elements, but no tie elements. Having these two types of information available can make interchange
    considerably easier, as some programs handle one type of information much more readily than the other.

    The print-leger attribute is used to indicate whether leger lines are printed. Notes without leger lines are
    used to indicate indeterminate high and low notes. By default, it is set to yes. If print-object is set to no,
    print-leger is interpreted to also be set to no if not present. This attribute is ignored for rests.

    The dynamics and end-dynamics attributes correspond to MIDI 1.0's Note On and Note Off velocities, respectively.
    They are expressed in terms of percentages of the default forte value (90 for MIDI 1.0).

    The attack and release attributes are used to alter the starting and stopping time of the note from when it
    would otherwise occur based on the flow of durations - information that is specific to a performance. They are
    expressed in terms of divisions, either positive or negative. A note that starts a tie should not have a
    release attribute, and a note that stops a tie should not have an attack attribute. The attack and release
    attributes are independent of each other. The attack attribute only changes the starting time of a note, and
    the release attribute only changes the stopping time of a note.

    If a note is played only particular times through a repeat, the time-only attribute shows which times to
    play the note.

    The pizzicato attribute is used when just this note is sounded pizzicato, vs. the pizzicato element which
    changes overall playback between pizzicato and arco.
    """
    _DTD = Sequence(
        Choice(
            Sequence(
                GroupReference(FullNote),
                GroupReference(DurationGroup),
                Element(Tie, 0, 2)
            ),
            Sequence(
                Element(Grace),
                Choice(
                    Sequence(
                        GroupReference(FullNote)
                    ),
                    Sequence(
                        GroupReference(FullNote),
                        Element(Tie, 0, 2)
                    ),
                    Sequence(
                        Element(Cue),
                        GroupReference(FullNote)
                    )
                )
            ),
            Sequence(
                Element(Cue),
                GroupReference(FullNote),
                GroupReference(DurationGroup)
            ),
        ),
        Element(Instrument, 0),
        GroupReference(EditorialVoice, 0),
        Element(Type, 0),
        Element(Dot, 0, None),
        Element(Accidental, 0),
        Element(TimeModification, 0, None),
        Element(Stem, 0),
        Element(Notehead, 0),
        Element(NoteheadText, 0),
        GroupReference(Staff, 0),
        Element(Beam, 0, 8),
        Element(Notations, 0, None),
        Element(Lyric, 0, None),
        Element(Play, 0)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='note', *args, **kwargs)



from quicktions import Fraction

from musicscore.musicxml.elements.fullnote import Rest, Event, Pitch, Alter, Chord
from musicscore.musicxml.elements.note import Note, Duration, Grace, Accidental, Notations, Lyric
from musicscore.musicxml.groups.musicdata import Backup
from musicscore.musicxml.types.complextypes.lyric import Text


class TreeBackup(Backup):
    def __init__(self, quarter_duration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quarter_duration = None
        self.quarter_duration = quarter_duration

    @property
    def quarter_duration(self):
        return self._quarter_duration

    @quarter_duration.setter
    def quarter_duration(self, value):
        self._quarter_duration = value

    def update_duration(self, divisions):
        try:
            self.get_children_by_type(Duration)[0].value = int(self.quarter_duration * divisions)
        except IndexError:
            self.add_child(Duration())
            self.duration.value = int(self.quarter_duration * divisions)


class TreeAccidental(Accidental):
    """"""

    def __init__(self, value='natural', show=False, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)
        self._note = None
        self._show = None
        self.show = show

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, value):
        if not isinstance(value, bool):
            raise TypeError('show.value must be of type bool not{}'.format(type(value)))
        self._show = value
        if self._note:
            self._note.update_accidental()


class TreeNote(Note):
    # todo: Inheritance Chain must be Midi, MidiNote, TreeMidiNote (or TreeNote)
    """
    quarter_duration = 0 means grace note
    """

    def __init__(self, event=Rest(), quarter_duration=1, is_tied=False, parent_chord=None, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.parent_chord = parent_chord
        self._accidental = TreeAccidental(show=False, value='natural')
        self._accidental._note = self
        self._quarter_duration = None
        self.quarter_duration = quarter_duration
        self._event = None
        self.event = event
        self._offset = None
        self._is_tied = False
        self.is_tied = is_tied
        self.is_finger_tremolo = False


    @property
    def quarter_duration(self):
        return self._quarter_duration

    @quarter_duration.setter
    def quarter_duration(self, value):
        if not isinstance(value, int) and not isinstance(value, float) and not isinstance(value, Fraction):
            raise TypeError('quarter_duration must be of type int, float or Fraction not {}'.format(type(value)))

        if value < 0:
            raise ValueError('quarter_duration {} must be zero or positive'.format(value))

        if not isinstance(self.quarter_duration, Fraction):
            self._quarter_duration = Fraction(value).limit_denominator(10000)
        else:
            self._quarter_duration = value

        if value == 0:
            if not self.get_children_by_type(Grace):
                children = self.get_children()[:]
                self.reset_dtd()
                for child in children:
                    if not isinstance(child, Duration):
                        self.add_child(child)
                    else:
                        del child
                self.add_child(Grace())
            # if not self.is_finger_tremolo:
            #
            # else:
            #     self._quarter_duration = 1

    @property
    def previous(self):
        index = self.up.notes.index(self)
        if index == 0:
            return None
        return self.up.notes[index - 1]

    def update_offset(self):
        if self.previous:
            if self.get_children_by_type(Chord):
                output = self.previous.offset
            else:
                output = self.previous.offset + self.previous.quarter_duration
            self._offset = output
        else:
            self._offset = 0

    @property
    def offset(self):
        if self._offset is None:
            self.update_offset()
        return self._offset

    @property
    def end_position(self):
        return self.offset + self.quarter_duration

    @property
    def accidental(self):
        return self._accidental

    @property
    def event(self):
        return self._event

    @event.setter
    def event(self, value):
        if not isinstance(value, Event):
            raise TypeError('event.value must be of type  not{}'.format(type(value)))
        try:
            self.remove_child(self._event)
        except ValueError as err:
            print(err)
            pass
        self._event = self.add_child(value)
        self.update_accidental()

    # @property
    # def is_finger_tremolo(self):
    #     try:
    #         return self.parent_chord.is_finger_tremolo
    #     except AttributeError:
    #         return False

    def update_accidental(self):
        _accidentals = {-1.5: 'three-quarters-flat',
                        -1: 'flat',
                        -0.5: 'quarter-flat',
                        0: 'natural',
                        0.5: 'quarter-sharp',
                        1: 'sharp',
                        1.5: 'three-quarters-sharp',
                        2: 'double-sharp'
                        }
        accidental = self.get_children_by_type(type_=TreeAccidental)
        if self._accidental.show:
            if isinstance(self.event, Pitch):
                alter = self.event.get_children_by_type(type_=Alter)
                if not alter:
                    self.event.add_child(Alter(0))

                accidental_value = _accidentals[self.event.alter.value]
                if accidental:
                    accidental[0].value = accidental_value
                else:
                    self._accidental.value = accidental_value
                    self.add_child(self._accidental)
        else:
            if accidental:
                self.remove_child(accidental[0])

    def _add_notations(self, notation):
        try:
            notations = self.get_children_by_type(Notations)[0]
        except IndexError:
            notations = self.add_child(Notations())

        notations.add_child(notation)

    def update_duration(self, divisions):
        try:
            self.duration.value = int(self.quarter_duration * divisions)
        except AttributeError:
            self.add_child(Duration())
            self.duration.value = int(self.quarter_duration * divisions)

    def add_lyric(self, text, number=1, **kwargs):
        lyric = self.add_child(Lyric(number=str(number), **kwargs))
        lyric.add_child(Text(str(text)))
        return lyric



"""
    <xs:complexType name="appearance">
        <xs:sequence>
            <xs:element name="line-width" type="line-width" minOccurs="0" maxOccurs="unbounded"/>
            <xs:element name="note-size" type="note-size" minOccurs="0" maxOccurs="unbounded"/>
            <xs:element name="distance" type="distance" minOccurs="0" maxOccurs="unbounded"/>
            <xs:element name="glyph" type="glyph" minOccurs="0" maxOccurs="unbounded"/>
            <xs:element name="other-appearance" type="other-appearance" minOccurs="0" maxOccurs="unbounded"/>
        </xs:sequence>
    </xs:complexType>
"""
from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.distance import ComplexTypeDistance
from musicscore.musicxml.types.complextypes.glyph import ComplexTypeGlyph
from musicscore.musicxml.types.complextypes.linewidth import ComplexTypeLineWidth
from musicscore.musicxml.types.complextypes.notesize import ComplexTypeNoteSize
from musicscore.musicxml.types.complextypes.otherappearance import ComplexTypeOtherAppearance


class LineWidth(ComplexTypeLineWidth):
    """"""

    _TAG = 'line-width'

    def __init__(self, type, *args, **kwargs):
        super().__init__(tag=self._TAG, type=type, *args, **kwargs)


class NoteSize(ComplexTypeNoteSize):
    """"""

    _TAG = 'note-size'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Distance(ComplexTypeDistance):
    """"""

    _TAG = 'distance'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Glyph(ComplexTypeGlyph):
    """"""

    _TAG = 'glyph'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class OtherAppearance(ComplexTypeOtherAppearance):
    """"""

    _TAG = 'other-appearance'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeAppearance(ComplexType):
    """The appearance type controls general graphical settings for the music's final form appearance on a printed page
    of display. This includes support for line widths, definitions for note sizes, and standard distances between
    notation elements, plus an extension element for other aspects of appearance."""

    _DTD = Sequence(
        Element(LineWidth, 0, None),
        Element(NoteSize, 0, None),
        Element(Distance, 0, None),
        Element(Glyph, 0, None),
        Element(OtherAppearance, 0, None)

    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.dtd.dtd import Choice, Element
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.types.complextypes.arrow import ComplexTypeArrow
from musicscore.musicxml.types.complextypes.bend import ComplexTypeBend
from musicscore.musicxml.types.complextypes.complextype import ComplexType, EmptyPlacement, EmptyPlacementSmulf, \
    PlacementText, OtherPlacementText
from musicscore.musicxml.types.complextypes.fingering import ComplexTypeFingering
from musicscore.musicxml.types.complextypes.fret import ComplexTypeFret
from musicscore.musicxml.types.complextypes.hammeronpulloff import ComplexTypeHammerOnPullOff
from musicscore.musicxml.types.complextypes.handbell import ComplexTypeHandbell
from musicscore.musicxml.types.complextypes.harmonic import ComplexTypeHarmonic
from musicscore.musicxml.types.complextypes.harmonmute import ComplexTypeHarmonMute
from musicscore.musicxml.types.complextypes.heeltoe import ComplexTypeHeelToe
from musicscore.musicxml.types.complextypes.hole import ComplexTypeHole
from musicscore.musicxml.types.complextypes.string_ import ComplexTypeString_
from musicscore.musicxml.types.complextypes.tap import ComplexTypeTap


class UpBow(EmptyPlacement):
    """
    The up-bow element represents the symbol that is used both for up-bowing on bowed instruments, and up-stroke on
    plucked instruments.
    """
    _TAG = 'up-bow'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class DownBow(EmptyPlacement):
    """
    The down-bow element represents the symbol that is used both for down-bowing on bowed instruments, and down-stroke
    on plucked instruments.
    """
    _TAG = 'down-bow'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Harmonic(ComplexTypeHarmonic):
    _TAG = 'harmonic'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class OpenString(EmptyPlacement):
    """
    The open-string element represents the zero-shaped open string symbol.
    """
    _TAG = 'open-string'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ThumbPosition(EmptyPlacement):
    """
    The thumb-position element represents the thumb position symbol. This is a circle with a line, where the line does
    not come within the circle. It is distinct from the snap pizzicato symbol, where the line comes inside the circle.
"""
    _TAG = 'thumb-position'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Fingering(ComplexTypeFingering):
    _TAG = 'fingering'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Pluck(PlacementText):
    """
    The pluck element is used to specify the plucking fingering on a fretted instrument, where the fingering element
    refers to the fretting fingering. Typical values are p, i, m, a for pulgar/thumb, indicio/index, medio/middle,
    and anular/ring fingers.
    """
    _TAG = 'placement-text'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class DoubleTongue(EmptyPlacement):
    """he double-tongue element represents the double tongue symbol (two dots arranged horizontally)."""

    _TAG = 'double-tongue'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class TripleTongue(EmptyPlacement):
    """The triple-tongue element represents the triple tongue symbol (three dots arranged horizontally)."""

    _TAG = 'triple-tongue'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Stopped(EmptyPlacementSmulf):
    """
    The stopped element represents the stopped symbol, which looks like a plus sign. The smufl attribute distinguishes
    different SMuFL glyphs that have a similar appearance such as handbellsMalletBellSuspended and guitarClosePedal.
    If not present, the default glyph is brassMuteClosed.
    """

    _TAG = 'stopped'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SnapPizzicato(EmptyPlacement):
    """The snap-pizzicato element represents the snap pizzicato symbol. This is a circle with a line, where the line
    comes inside the circle. It is distinct from the thumb-position symbol, where the line does not come inside the
    circle."""

    _TAG = 'snap-pizzicato'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Fret(ComplexTypeFret):
    """"""

    _TAG = 'fret'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class String_(ComplexTypeString_):
    """"""

    _TAG = 'string'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class HammerOn(ComplexTypeHammerOnPullOff):
    """"""

    _TAG = 'hammer-on'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PullOff(ComplexTypeHammerOnPullOff):
    """"""

    _TAG = 'pull-off'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Bend(ComplexTypeBend):
    """"""

    _TAG = 'bend'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Tap(ComplexTypeTap):
    """"""

    _TAG = 'tap'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Heel(ComplexTypeHeelToe):
    """"""

    _TAG = 'heel'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Toe(ComplexTypeHeelToe):
    """"""

    _TAG = 'toe'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Fingernails(EmptyPlacement):
    """The fingernails element is used in notation for harp and other plucked string instruments."""

    _TAG = 'fingernails'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Hole(ComplexTypeHole):
    """"""

    _TAG = 'hole'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Arrow(ComplexTypeArrow):
    """"""

    _TAG = 'arrow'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Handbell(ComplexTypeHandbell):
    """"""

    _TAG = 'handbell'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class BrassBend(EmptyPlacement):
    """The brass-bend element represents the u-shaped bend symbol used in brass notation,  distinct from the bend
    element used in guitar music."""

    _TAG = 'brass-bend'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Flip(EmptyPlacement):
    """The flip element represents the flip symbol used in brass notation."""

    _TAG = 'flip'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Smear(EmptyPlacement):
    """The smear element represents the tilde-shaped smear symbol used in brass notation."""

    _TAG = 'smear'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Open(EmptyPlacementSmulf):
    """The open element represents the open symbol, which looks like a circle. The smufl attribute can be used to
    distinguish different SMuFL glyphs that have a similar appearance such as brassMuteOpen and guitarOpenPedal. If not
    present, the default glyph is brassMuteOp
    """
    _TAG = 'open'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class HalfMuted(EmptyPlacementSmulf):
    """The half-muted element represents the half-muted symbol, which looks like a circle with a plus sign inside.
    The smufl attribute can be used to distinguish different SMuFL glyphs that have a similar appearance such as
    brassMuteHalfClosed and guitarHalfOpenPedal. If not present, the default glyph is brassMuteHalfClosed."""

    _TAG = 'half-muted'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class HarmonMute(ComplexTypeHarmonMute):
    """"""

    _TAG = 'harmon-mute'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Golpe(EmptyPlacement):
    """The golpe element represents the golpe symbol that is used for tapping the pick guard in guitar music."""

    _TAG = 'golpe'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class OtherTechnical(OtherPlacementText):
    """The other-technical element is used to define any technical indications not yet in the MusicXML format. The smufl
    attribute can be used to specify a particular glyph, allowing application interoperability without requiring every
    SMuFL technical indication to have a MusicXML element equivalent. Using the other-technical element without the smufl
    attribute allows for extended representation, though without application interoperability."""

    _TAG = 'other-technical'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeTechnical(ComplexType, OptionalUniqueId):
    """
    Technical indications give performance information for individual instruments.
    """
    _DTD = Choice(
        Element(UpBow),
        Element(DownBow),
        Element(Harmonic),
        Element(OpenString),
        Element(ThumbPosition),
        Element(Fingering),
        Element(Pluck),
        Element(DoubleTongue),
        Element(TripleTongue),
        Element(Stopped),
        Element(SnapPizzicato),
        Element(Fret),
        Element(String_),
        Element(HammerOn),
        Element(PullOff),
        Element(Bend),
        Element(Tap),
        Element(Heel),
        Element(Toe),
        Element(Fingernails),
        Element(Hole),
        Element(Arrow),
        Element(Handbell),
        Element(BrassBend),
        Element(Flip),
        Element(Smear),
        Element(Open),
        Element(HalfMuted),
        Element(HarmonMute),
        Element(Golpe),
        Element(OtherTechnical),

        min_occurrence=0,
        max_occurrence=None)

    def __init__(self, tag, *args, **kwargs):
        super().__init__(*args, tag=tag, **kwargs)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Notations, Duration
from musicscore.musicxml.types.complextypes.dynamics import FF, Dynamics
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        chord = TreeChord()
        notations = chord.add_child(Notations())
        dynamics = notations.add_child(Dynamics())
        dynamics.add_child(FF())
        note = chord._notes[0]
        note.add_child(Duration(1))
        result = '''<note>
  <pitch>
    <step>B</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <notations>
    <dynamics placement="below">
      <ff/>
    </dynamics>
  </notations>
</note>
'''
        self.assertEqual(note.to_string(), result)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[1, 1])

        sf.chords[0].add_dynamics('fff')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

#         result = '''<note>
#   <pitch>
#     <step>B</step>
#     <octave>4</octave>
#   </pitch>
#   <duration>1</duration>
#   <notations>
#     <dynamics placement="below">
#       <ff/>
#     </dynamics>
#   </notations>
# </note>
# '''
#         self.assertEqual(note.to_string(), result)

#     def test_2(self):
#         chord = TreeChord()
#         chord.add_dynamics('fp')
#         chord.add_dynamics('pp')
#         note = chord._notes[0]
#         note.add_child(Duration(1))
#         result = '''<note>
#   <pitch>
#     <step>B</step>
#     <octave>4</octave>
#   </pitch>
#   <duration>1</duration>
#   <notations>
#     <dynamics placement="below">
#       <fp/>
#     </dynamics>
#     <dynamics placement="below">
#       <pp/>
#     </dynamics>
#   </notations>
# </note>
# '''
#         self.assertEqual(note.to_string(), result)
#
#         with self.assertRaises(ValueError):
#             chord.add_dynamics('mmf')
#
#     def test_3(self):
#         chord = TreeChord()
#         d = chord.add_dynamics('pp')
#         d.relative_x = 10
#         d.valign = 'middle'
#         d.halign = 'center'
#         note = chord._notes[0]
#         note.add_child(Duration(1))
#
#         result = '''<note>
#   <pitch>
#     <step>B</step>
#     <octave>4</octave>
#   </pitch>
#   <duration>1</duration>
#   <notations>
#     <dynamics relative-x="10" halign="center" valign="middle" placement="below">
#       <pp/>
#     </dynamics>
#   </notations>
# </note>
# '''
#         self.assertEqual(note.to_string(), result)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.holeclosed import ComplexTypeHoleClosed
from musicscore.musicxml.types.simple_type import String


class HoleType(String):
    """The content of the optional hole-type element indicates what the hole symbol represents in terms of instrument
    fingering or other techniques."""

    _TAG = 'hole-type'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class HoleClosed(ComplexTypeHoleClosed):
    """"""

    _TAG = 'hole-closed'

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class HoleShape(String):
    """The optional hole-shape element indicates the shape of the hole symbol; the default is a circle."""

    _TAG = 'hole-shape'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeHole(ComplexType, PrintStyle, Placement):
    """The hole type represents the symbols used for woodwind and brass fingerings as well as other notations."""
    _DTD = Sequence(
        Element(HoleType, min_occurrence=0),
        Element(HoleClosed),
        Element(HoleShape, min_occurrence=0)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.score.add_measure()
        self.score.add_part()
        self.score.add_chord(1, 1, TreeChord())
        self.score.add_chord(1, 1, TreeChord(quarter_duration=0))
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[0.75, 0.25, 1, 0.5, 4])
        sf.chords[1].set_manual_type('16th')
        sf.chords[1].quarter_duration = 0
        sf.chords[0].set_manual_type('eighth')
        sf.chords[0].set_manual_dots(1)
        sf.chords[0].quarter_duration = 0
        sf.chords[2].set_manual_type('quarter')
        sf.chords[2].quarter_duration = 0
        sf.chords[3].set_manual_type('eighth')
        sf.chords[3].quarter_duration = 0

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + 'test_2.xml'
        self.score.write(xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_3(self):
        sf = SimpleFormat(quarter_durations=[4])
        sf.chords[0].add_grace_chords([TreeChord(60), TreeChord(63), TreeChord(68)])
        sf.chords[0].add_grace_chords([TreeChord(61), TreeChord(66)], mode='post')
        print(sf.chords[0].get_pre_grace_chords())
        print(sf.chords[0].get_post_grace_chords())

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + 'test_3.xml'
        self.score.write(xml_path)
        # TestScore().assert_template(result_path=xml_path)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.credit import ComplexTypeCredit
from musicscore.musicxml.types.complextypes.defaults import ComplexTypeDefaults
from musicscore.musicxml.types.complextypes.partlist import ComplexTypePartList
from musicscore.musicxml.types.complextypes.scorepart import Identification


class Work(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='work', *args, **kwargs)
        raise NotImplementedError('Work')


class MovementNumber(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='movement-number', *args, **kwargs)
        raise NotImplementedError('MovementNumber')


class MovementTitle(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='movement-title', *args, **kwargs)
        raise NotImplementedError('MovementTitle')


class Defaults(ComplexTypeDefaults):
    _TAG = 'defaults'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Credit(ComplexTypeCredit):
    _TAG = 'credit'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PartList(ComplexTypePartList):
    _TAG = 'part-list'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


"""
The score-header group contains basic score metadata about the work and movement, score-wide defaults for layout and 
fonts, credits that appear on the first or following pages, and the part list.
"""
ScoreHeader = Sequence(
    Element(Work, min_occurrence=0),
    Element(MovementNumber, min_occurrence=0),
    Element(MovementTitle, min_occurrence=0),
    Element(Identification, min_occurrence=0),
    Element(Defaults, min_occurrence=0),
    Element(Credit, min_occurrence=0, max_occurrence=None),
    Element(PartList)
)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, NonNegativeInteger, TypeNoteTypeValue
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.lineshape import LineShape
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeNumberLevel, TypeStartStop


class TupletDot(XMLElement, Font, Color):
    """
    The tuplet-dot type is used to specify dotted normal tuplet types.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(tag='tuplet-dot', *args, **kwargs)


class TupletNumber(XMLElement, NonNegativeInteger, Font, Color):
    """
    The tuplet-number type indicates the number of notes for this portion of the tuplet.
    """

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag='tuplet-number', value=value, *args, **kwargs)


class TupletType(XMLElement, TypeNoteTypeValue, Font, Color):
    """
	The tuplet-type type indicates the graphical note type of the notes for this portion of the tuplet.</xs:documentation>
    """

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag='tuplet-type', value=value, *args, **kwargs)


class ComplexTypeTupletPortion(ComplexType):
    """
    The tuplet-portion type provides optional full control over tuplet specifications. It allows the number and note
    type (including dots) to be set for the actual and normal portions of a single tuplet. If any of these elements are
    absent, their values are based on the time-modification element.
    """
    _DTD = Sequence(
        Element(TupletNumber, min_occurrence=0),
        Element(TupletType, min_occurrence=0),
        Element(TupletDot, min_occurrence=0, max_occurrence=None)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class TupletActual(ComplexTypeTupletPortion):
    """
    The tuplet-actual element provide optional full control over how the actual part of the tuplet is displayed,
    including number and note type (with dots). If any of these elements are absent, their values are based on the
    time-modification element.
    """

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag='tuplet-actual', value=value, *args, **kwargs)


class TupletNormal(ComplexTypeTupletPortion):
    """
    The tuplet-normal element provide optional full control over how the normal part of the tuplet is displayed,
    including number and note type (with dots). If any of these elements are absent, their values are based on the
    time-modification element.
    """

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag='tuplet-normal', value=value, *args, **kwargs)


class Bracket(AttributeAbstract):
    def __init__(self, bracket=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bracket', bracket, 'TypeYesNo')


class ShowNumber(AttributeAbstract):
    def __init__(self, show_number=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('show-number', show_number, 'TypeShowTuplet')


class ShowType(AttributeAbstract):
    def __init__(self, show_type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('show-type', show_type, 'TypeShowTuplet')


class ComplexTypeTuplet(ComplexType, Bracket, ShowNumber, ShowType, LineShape, Position, Placement, OptionalUniqueId):
    """
    A tuplet element is present when a tuplet is to be displayed graphically, in addition to the sound data provided by
    the time-modification elements. The number attribute is used to distinguish nested tuplets. The bracket attribute is
    used to indicate the presence of a bracket. If unspecified, the results are implementation-dependent. The line-shape
    attribute is used to specify whether the bracket is straight or in the older curved or slurred style. It is straight
    by default.

    Whereas a time-modification element shows how the cumulative, sounding effect of tuplets and double-note tremolos
    compare to the written note type, the tuplet element describes how this is displayed. The tuplet element also
    provides more detailed representation information than the time-modification element, and is needed to represent
    nested tuplets and other complex tuplets accurately.

    The show-number attribute is used to display either the number of actual notes, the number of both actual and normal
    notes, or neither. It is actual by default. The show-type attribute is used to display either the actual type, both
    the actual and normal types, or neither. It is none by default.
    """
    _DTD = Sequence(
        Element(TupletActual, min_occurrence=0),
        Element(TupletNormal, min_occurrence=0)
    )

    def __init__(self, type, number=1, *args, **kwargs):
        super().__init__(tag='tuplet', *args, **kwargs)
        self.type = type
        self.number = number

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeNumberLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStop(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat, TreeChord
from musicscore.musictree.treechordflags import FingerTremoloFlag, PercussionFlag
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

from musicscore.musicxml.elements.note import TimeModification, Stem
from musicscore.musicxml.types.complextypes.timemodification import ActualNotes, NormalNotes

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + "_test_1.xml"
        sf = SimpleFormat(midis=[60, 63], quarter_durations=[0.5, 0.5, 3, 4, 4, 4])

        sf.chords[0].set_manual_type('quarter')
        sf.chords[1].set_manual_type('quarter')

        sf.chords[1].add_child(Stem('none'))
        sf.chords[0].quarter_duration *= 2
        # sf.chords[1].add_child(Stem('up'))

        # sf.chords[0].add_bracket(type='start', line_end='down', relative_x=0)
        # sf.chords[2].add_bracket(type='stop', line_end='none')

        sf.chords[0].add_words('\uE227', font_family='bravura', font_size=16, relative_x=30, relative_y=-50)

        tm = TimeModification()
        tm.add_child(ActualNotes(0))
        tm.add_child(NormalNotes(1))
        sf.chords[1].add_child(tm)

        sf.to_stream_voice().add_to_score(self.score)

        self.score.write(xml_path)




from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import String


class Substitution(AttributeAbstract):
    """"""

    def __init__(self, substitution=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('substitution', substitution, 'TypeYesNo')


class Alternate(AttributeAbstract):
    """"""

    def __init__(self, alternate=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('alternate', alternate, 'TypeYesNo')


class ComplexTypeFingering(ComplexType, String, Substitution, Alternate, PrintStyle, Placement):
    """
    Fingering is typically indicated 1,2,3,4,5. Multiple fingerings may be given, typically to substitute fingerings in
    the middle of a note. The substitution and alternate values are "no" if the attribute is not present. For guitar
    and other fretted instruments, the fingering element represents the fretting finger; the pluck element represents
    the plucking finger.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from unittest import TestCase
import os

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.score.add_measure(TreeMeasure(time=(5, 8)))
        # sf = SimpleFormat(durations=[1.8, 0.2, 0.5])
        sf = SimpleFormat(quarter_durations=[2.5])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_1'
        self.score.finish()
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        self.score.add_measure(TreeMeasure(time=(5, 8)))
        # sf = SimpleFormat(durations=[1.8, 0.2, 0.5])
        sf = SimpleFormat(
            quarter_durations=[Fraction(1, 7), Fraction(6, 7), Fraction(1, 14), Fraction(3, 14), Fraction(10, 14),
                               Fraction(1, 2)])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 2)
        self.score.get_score_parts()[1].max_division = 7

        result_path = path + '_test_2'
        self.score.finish()
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)


'''
    <xs:attributeGroup name="print-attributes">
        <xs:annotation>

            </xs:documentation>
        </xs:annotation>
        <xs:attribute name="staff-spacing" type="tenths"/>
        <xs:attribute name="new-system" type="yes-no"/>
        <xs:attribute name="new-page" type="yes-no"/>
        <xs:attribute name="blank-page" type="xs:positiveInteger"/>
        <xs:attribute name="page-number" type="xs:token"/>
    </xs:attributeGroup>
'''

from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class StaffSpacing(AttributeAbstract):
    """"""

    def __init__(self, staff_spacing=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('staff-spacing', staff_spacing, "TypeTenths")


class NewSystem(AttributeAbstract):
    """"""

    def __init__(self, new_system=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('new-system', new_system, "TypeYesNo")


class NewPage(AttributeAbstract):
    """"""

    def __init__(self, new_page=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('new-page', new_page, "TypeYesNo")


class BlankPage(AttributeAbstract):
    """"""

    def __init__(self, blank_page=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('blank-page', blank_page, "PositiveInteger")


class PageNumber(AttributeAbstract):
    """"""

    def __init__(self, page_number=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('page-number', page_number, "Token")



class PrintAttributes(StaffSpacing, NewSystem, NewPage, BlankPage, PageNumber):
    """
    The print-attributes group is used by the print element. The new-system and new-page attributes indicate whether to 
    force a system or page break, or to force the current music onto the same system or page as the preceding music. 
    Normally this is the first music data within a measure. If used in multi-part music, they should be placed in the 
    same positions within each part, or the results are undefined. The page-number attribute sets the number of a new 
    page; it is ignored if new-page is not "yes". Version 2.0 adds a blank-page attribute. This is a positive integer 
    value that specifies the number of blank pages to insert before the current measure. It is ignored if new-page is 
    not "yes". These blank pages have no music, but may have text or images specified by the credit element. This is 
    used to allow a combination of pages that are all text, or all text and images, together with pages of music.

    The staff-spacing attribute specifies spacing between multiple staves in tenths of staff space. This is deprecated 
    as of Version 1.1; the staff-layout element should be used instead. If both are present, the staff-layout values 
    take priority.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[1, 1, 1])

        sf.chords[0].add_fermata()
        sf.chords[1].add_fermata('square')
        sf.chords[2].add_fermata('angled')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)



from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.fullnote import Rest
from musicscore.musicxml.elements.note import Note, Duration
from musicscore.musicxml.elements.scoreheader import PartList
from musicscore.musicxml.groups.musicdata import Attributes
from musicscore.musicxml.types.complextypes.attributes import Clef, Divisions, Time, Beats, BeatType
from musicscore.musicxml.types.complextypes.clef import Sign, Line
from musicscore.musicxml.types.complextypes.partlist import ScorePart
from musicscore.musicxml.types.complextypes.scorepart import PartName, PartAbbreviation
from musicscore.musicxml.elements.timewise import Score, Measure, Part

import os

from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class TestTimewise(TestScore):
    def setUp(self):
        self.score = Score()
        self.score.version = '3.0'
        part_list = self.score.add_child(PartList())
        score_part = part_list.add_child(ScorePart(id='P1'))
        score_part.add_child(PartName(name="Oboe"))
        score_part.add_child(PartAbbreviation("Ob."))

        measure = self.score.add_child(Measure(number='1'))
        part = measure.add_child(Part(id='P1'))
        attributes = part.add_child(Attributes())
        attributes.add_child(Divisions(2))
        time = attributes.add_child(Time())
        time.add_child(Beats(4))
        time.add_child(BeatType(4))
        clef = attributes.add_child(Clef())
        clef.add_child(Sign('G'))
        clef.add_child(Line(2))
        note = part.add_child(Note())
        note.add_child(Rest(measure='yes'))
        note.add_child(Duration(8))

        measure = self.score.add_child(Measure(number='2'))
        part = measure.add_child(Part(id='P1'))
        attributes = part.add_child(Attributes())
        attributes.add_child(Divisions(2))
        note = part.add_child(Note())
        note.add_child(Rest(measure='yes'))
        note.add_child(Duration(8))

    def test_score_head(self):
        # print(self.score.to_string())
        self.score.write(path=path)
        self.assert_template(path)




from musicscore.musictree.midi import Midi
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treeclef import TREBLE_CLEF, BASS_CLEF, LOW_BASS_CLEF, HIGH_TREBLE_CLEF
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart
from musicscore.musicxml.groups.common import Voice
from musicscore.musicxml.types.simple_type import PositiveInteger


class StreamVoice(object):

    def __init__(self, voice_number=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._chords = []
        self._voice = None
        self.voice_number = voice_number

    @property
    def chords(self):
        return self._chords

    @property
    def voice_number(self):
        return self._voice

    @voice_number.setter
    def voice_number(self, value):
        PositiveInteger(value)
        self._voice = value

    def add_chord(self, chord):
        chord.voice_number = self.voice_number
        self._chords.append(chord)

    def add_to_part(self, part, chords=None):
        if chords is None:
            chords = self.chords

        for i in range(len(chords)):
            chord = chords[i]
            remain = part.add_chord(chord, self.voice_number)
            if remain:
                remaining_chords = [remain] + chords[i + 1:]
                return remaining_chords

    def add_to_score(self, score, first_measure=1, part_number=1):
        measure_number = first_measure

        def _get_measure():
            try:
                measure = score.get_children_by_type(TreeMeasure)[measure_number - 1]
            except IndexError:
                measure = score.add_measure()

            return measure

        measure = _get_measure()
        for i in range(part_number - len(measure.get_children_by_type(TreePart))):
            score.add_part()

        def _add_to_part(chords=None):
            part = measure.get_children_by_type(TreePart)[part_number - 1]
            remain = self.add_to_part(part, chords)
            return remain

        remain = _add_to_part()

        while remain:
            measure_number += 1
            measure = _get_measure()
            remain = _add_to_part(remain)


class StreamChordFormula(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def condition(self, chord):
        pass

    def change_chord(self, chord):
        pass


class SimpleFormat(object):
    def __init__(self, quarter_durations=None, midis=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quarter_durations = None
        self._midis = None
        self._set_quarter_durations(quarter_durations)
        self._set_midis(midis)
        self._chords = None

    @property
    def quarter_duration(self):
        return sum([chord.quarter_duration for chord in self.chords])

    @property
    def midis(self):
        return [chord.midis for chord in self.chords]

    @property
    def chords(self):
        if self._chords is None:
            self._generate_chords()
        return self._chords

    @staticmethod
    def _check_quarter_duration(quarter_durations):
        for quarter_duration in quarter_durations:
            if quarter_duration < 0:
                raise ValueError('SimpleFormat(): wrong duration {}'.format(quarter_duration))

    def _set_quarter_durations(self, quarter_durations):
        if quarter_durations is None:
            quarter_durations = []
        else:
            try:
                quarter_durations = list(quarter_durations)
            except TypeError:
                quarter_durations = [quarter_durations]

        self._check_quarter_duration(quarter_durations)

        self._quarter_durations = quarter_durations

    def _set_midis(self, midis):
        def _is_value(x):
            try:
                float(x)
                return True
            except (ValueError, TypeError):
                return False

        def _is_midi(x):
            if isinstance(x, Midi):
                return True
            else:
                return False

        if midis is None:
            midis = []
        else:
            try:
                midis = list(midis)
            except TypeError:
                midis = [midis]

        self._midis = []
        for m in midis:
            if _is_midi(m):
                self._midis.append(m)
            elif _is_value(m):
                self._midis.append(Midi(m))
            elif isinstance(m, list) or isinstance(m, tuple):
                tmp = []

                for midi in m:
                    if _is_midi(midi):
                        tmp.append(midi)
                    elif _is_value(midi):
                        tmp.append(Midi(midi))
                    else:
                        raise TypeError
                self._midis.append(tmp)
            else:
                raise TypeError

    def get_quarter_durations(self):
        return [chord.quarter_duration for chord in self.chords]

    def get_midis(self):
        return [chord.midis for chord in self.chords]

    def _generate_chords(self):
        if self._quarter_durations == [] and self._midis == []:
            pass
        else:
            self._chords = []
            for i in range(len(self._quarter_durations) - len(self._midis)):
                self._midis.append(71)

            for i in range(len(self._midis) - len(self._quarter_durations)):
                self._quarter_durations.append(1)

            for d, m in zip(self._quarter_durations, self._midis):
                self._chords.append(TreeChord(m, d))

    def add_chord(self, chord):
        if self._chords is None:
            self._chords = []
        if chord.get_children_by_type(Voice):
            raise Exception('SimpleFormat Chords cannot have a voice child.')
        self._chords.append(chord)
        return chord

    def to_stream_voice(self, voice_number=1):
        output = StreamVoice(voice_number)
        for chord in self.chords:
            output.add_chord(chord.__deepcopy__())
            # output.add_chord(chord)
        return output

    def __deepcopy__(self, memodict={}):
        output = SimpleFormat()
        for chord in self.chords:
            output.add_chord(chord.deepcopy_for_SimpleFormat())
        return output

    def auto_clef(self, clefs=None):
        if not clefs:
            clefs = [TREBLE_CLEF, BASS_CLEF, HIGH_TREBLE_CLEF, LOW_BASS_CLEF]
        current_clefs = []
        chords = self.chords

        def _get_possible_clefs(midi):
            output = []
            for clef in clefs:
                clef_in_range = False
                if None not in clef.optimal_range:
                    if clef.optimal_range[0] <= midi <= clef.optimal_range[1]:
                        clef_in_range = True
                elif clef.optimal_range[0] and clef.optimal_range[0] <= midi:
                    clef_in_range = True
                elif clef.optimal_range[1] and midi <= clef.optimal_range[1]:
                    clef_in_range = True
                if clef_in_range:
                    output.append(clef)
            return output

        for chord in [chord for chord in chords if not chord.is_rest]:
            try:
                last_clef = current_clefs[-1]
            except IndexError:
                last_clef = None
            midi = chord.midis[0]
            possible_clefs = _get_possible_clefs(midi)
            if not possible_clefs:
                raise ValueError(
                    'no possible clef war found for midi {} with value {}. Change clefs or optimal_range '
                    'of clefs.'.format(midi, midi.value))
            if not last_clef or last_clef not in possible_clefs:
                clef = possible_clefs[0]
            else:
                clef = None

            if clef and clef != last_clef:
                chord.add_clef(clef)
                current_clefs.append(clef)

    def change_chord(self, *chord_formulas):
        for formula in chord_formulas:
            if not isinstance(formula, StreamChordFormula):
                raise TypeError

        for chord in self.chords:
            for formula in chord_formulas:
                if formula.condition(chord):
                    formula.change_chord(chord)
                    break

    def extend(self, simple_format):
        # self.chords
        for chord in simple_format.chords:
            self.add_chord(chord)

    def transpose(self, interval):
        for ch in self.chords:
            ch.transpose(interval)



from unittest import TestCase
import os

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[7])
        v = sf.to_stream_voice()
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_1'
        self.score.write(path=result_path)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[12.5])
        v = sf.to_stream_voice()
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_2'
        self.score.write(path=result_path)

    def test_3(self):
        # sf = SimpleFormat(durations=[Fraction(4) + Fraction(1, 2), Fraction(1, 2)])
        sf = SimpleFormat(quarter_durations=[Fraction(4, 1) + Fraction(5, 7), Fraction(5, 7)])
        # sf = SimpleFormat(durations=[Fraction(4, 1) + Fraction(5, 7), Fraction(2, 7)])
        v = sf.to_stream_voice()
        v.add_to_score(self.score, 1, 1)

        self.score.finish()
        # print([chord.quarter_duration for chord in self.score.get_measure(2).get_part(1).chords])
        # print([chord.tie_types for chord in self.score.get_measure(2).get_part(1).chords])
        result_path = path + '_test_3'
        self.score.write(path=result_path)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class PrintObject(AttributeAbstract):
    """"""

    def __init__(self, print_object=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('print-object', print_object, "TypeYesNo")



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat, TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Notehead
from musicscore.musicxml.types.simple_type import TypeNoteheadValue
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[1, 2, 3, 2, 1])
        sf.chords[1].add_child(Notehead('square'))
        sf.chords[2].add_child(Notehead('diamond'))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        sf = SimpleFormat(midis=[(60, 62, 63)], quarter_durations=[4])
        sf.chords[0].add_child(Notehead('diamond'))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_3(self):
        sf = SimpleFormat(midis=[(60, 62, 63)], quarter_durations=[4])
        sf.chords[0].midis[0].notehead = Notehead('diamond')
        sf.chords[0].midis[1].notehead = Notehead('square')
        sf.chords[0].midis[2].notehead = Notehead('normal')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_3.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_4(self):
        sf = SimpleFormat()
        for notehead in TypeNoteheadValue._PERMITTED:
            sf.add_chord(TreeChord(quarter_duration=1))
            sf.chords[-1].midis[0].notehead = Notehead(notehead)
            sf.chords[-1].add_words(notehead)
            sf.add_chord(TreeChord(quarter_duration=2))
            sf.chords[-1].midis[0].notehead = Notehead(notehead)
        self.score.set_time_signatures(times={1: (3, 4)})
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_4.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)



# import cProfile
# import os
#
# from musicscore.musictree.treechord import TreeChord
# from musicscore.musictree.treescoretimewise import TreeScoreTimewise
#
# path = os.path.abspath(__file__).split('.')[0]
#
#
# def p():
#     score = TreeScoreTimewise()
#     score.add_measure()
#     score.add_part('one')
#     midis = [60, 61, 62, 60, 63, 64, 65, 61]
#     for midi in midis:
#         score.add_chord(1, 1, TreeChord(midi, quarter_duration=0.5))
#
#     score.finish()
#     score.write(path=path)
#
#
# cProfile.run('p()', sort="tottime")



import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])

score = TreeScoreTimewise()
score.page_style.orientation = 'landscape'
score.add_title('BIG IMPORTANT TITLE')
score.add_subtitle('unimportant subtitle')
score.add_composer('me')
score.add_text('some text')

simple_format = SimpleFormat(quarter_durations=4)
simple_format.to_stream_voice().add_to_score(score=score)

score.page_style.orientation = 'portrait'
# conversion of default_x and default_y in CreditWords (child of Credit)
xml_path = path + '_1.xml'
score.write(xml_path)



'''
	<xs:complexType name="other-text">
		<xs:annotation>
			<xs:documentation></xs:documentation>
		</xs:annotation>
		<xs:simpleContent>
			<xs:extension base="xs:string">
				<xs:attributeGroup ref="smufl"/>
			</xs:extension>
		</xs:simpleContent>
	</xs:complexType>
'''
from musicscore.musicxml.types.complextypes.complextype import ComplexType


class ComplexTypeOtherText(ComplexType):
    """
    The other-text type represents a text element with a smufl attribute group. This type is used by MusicXML direction
    extension elements to allow specification of specific SMuFL glyphs without needed to add every glyph as a MusicXML
    element.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        NotImplementedError()



from unittest import TestCase

from musicscore.musictree.midi import Midi
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treenote import TreeNote
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

from musicscore.musicxml.elements.note import Beam

path = os.path.abspath(__file__).split('.')[0]


class TestGrouping(TestCase):
    def setUp(self):
        self.score = TreeScoreTimewise()
        self.score.add_part()

    def test_grouping(self):
        self.make_measure(1, (3, 4))
        self.make_measure(2, (6, 8))

        # for measure in self.score.get_children_by_type(TreeMeasure):
        #     measure.get_part(1).group_beams()

        self.score.finish()

        self.score.write(path=path)

    def make_measure(self, number, time_signature=(4, 4)):
        self.score.add_measure(TreeMeasure(time=time_signature))
        for i in range(time_signature[0] * 8//time_signature[1]):
            self.score.add_chord(number, 1, TreeChord(60, quarter_duration=0.5))
